
from random import randint

import threed_strudel.utils.bioUtils
# from strudel.chop.chopMap import ChopMap, MapParser
from threed_strudel.parse.mapParser import MapParser
import math
import numpy as np
import time
from numpy import linalg as LA
import os
from scipy.interpolate import InterpolatedUnivariateSpline
from threed_strudel.utils import modelMapUtils


class ExtractDensity:
    def __init__(self, map_path=None, model_path=None, adapted_map_path=None, work_voxel_size=0.5, tmp_dir='/tmp/segment_density'):
        self.in_map = MapParser(map_path)
        self.cube_map = None
        self.box_size = 10
        self.max_thld = None
        self.min_thld = None
        self.in_model = threed_strudel.utils.bioUtils.load_structure(model_path)
        self.walk_grid = None
        self.operations_counter = 0
        self.max_roll = 5
        self.work_voxel_size = work_voxel_size
        self.interpolator = None
        self.tmp_dir = tmp_dir
        if not os.path.exists(tmp_dir):
            os.makedirs(tmp_dir)
        if adapted_map_path is not None:
            self.adapted_map = MapParser(adapted_map_path)
        else:
            self.adapted_map = None

    @staticmethod
    def cut_cube_around_voxel(map_obj, center, radius):
        radius = int(round(radius / map_obj.voxel_size[0]))
        size = radius * 2 + 1
        new_grid = np.zeros((size, size, size), dtype='float32')
        # Assign voxel values
        for x in range(size):
            for y in range(size):
                for z in range(size):
                    try:
                        new_grid[x, y, z] = map_obj.data[x + center[0] - radius,
                                                         y + center[1] - radius,
                                                         z + center[2] - radius]
                    except IndexError:
                        pass

        out_map = MapParser('')
        out_map.data = new_grid
        out_map.cell = (round(size * map_obj.voxel_size[0], 3),
                        round(size * map_obj.voxel_size[1], 3),
                        round(size * map_obj.voxel_size[2], 3))
        out_map.voxel_size = map_obj.voxel_size
        out_map.n_xyz = new_grid.shape
        n_st = map_obj.n_start
        out_map.n_start = (0, 0, 0)
        out_map.origin = ((center[2] - radius + n_st[0]) * out_map.voxel_size[0] + map_obj.origin[0],
                          (center[1] - radius + n_st[1]) * out_map.voxel_size[1] + map_obj.origin[1],
                          (center[0] - radius + n_st[2]) * out_map.voxel_size[2] + map_obj.origin[2])
        return out_map

    # def map_resample(self, map_obj, file_prefix=''):
    #     """
    #     Resample the grid in the a map object so that the
    #     voxel size become equal to 'self.work_voxel_size'
    #     :param map_obj: map object
    #     :param file_prefix: prefix for the tmp file
    #     :return: map object
    #     """
    #     tmp = os.path.join(self.tmp_dir, file_prefix + 'tmp.mrc')
    #     res_tmp = os.path.join(self.tmp_dir, file_prefix + 'res_tmp.mrc')
    #
    #
    #     map_obj.write_map(tmp)
    #     chop = ChopMap()
    #     chop.grid_resample(tmp, res_tmp, self.work_voxel_size, map_obj.voxel_size[0])
    #     out_map_obj = MapParser(res_tmp)
    #
    #     return out_map_obj

    def update_map_parameters(self, map_obj):
        self.max_thld = np.amax(map_obj.data)
        self.min_thld = np.amin(map_obj.data)

        self.walk_grid = np.zeros(map_obj.data.shape, dtype='float32')

    def get_flank_ca_coord(self, res):
        try:
            prev_ca = self.in_model[0][res.parent.id][res.id[1]-1]['CA'].coord
        except KeyError:
            prev_ca = None
        try:
            next_ca = self.in_model[0][res.parent.id][res.id[1]-1]['CA'].coord
        except KeyError:
            next_ca = None

        return prev_ca, next_ca

    def segment_residue_density(self, model):

        c_coord, ca_coord, n_coord, cb_coord = self._get_bb_coordinates(model)

        ca_index = self.coord_to_index_int(ca_coord, self.in_map)
        ca_map = self.cut_cube_around_voxel(self.in_map, ca_index, self.box_size)
        # print(ca_map.data.shape)
        res_nr = model.id[1]
        prev_ca_coord, next_ca_coord = self.get_flank_ca_coord(model)

        # self.cube_map = self.map_resample(ca_map, str(res_nr))
        ca_map.grid_resample_emda(self.work_voxel_size)
        self.cube_map = ca_map
        # print(self.cube_map.data.shape)

        self.update_map_parameters(self.cube_map)

        if self.adapted_map is not None:
            adapted_ca_map = self.cut_cube_around_voxel(self.adapted_map, ca_index, self.box_size)
            adapted_cube_map = adapted_ca_map.grid_resample_emda(self.work_voxel_size)
            self.interpolator = modelMapUtils.interpolator(adapted_cube_map)
        else:
            self.interpolator = modelMapUtils.interpolator(self.cube_map)



        start = time.time()
        if cb_coord != []:
            start_coord = cb_coord
        else:
            start_coord = ca_coord
        #start_coord = ca_coord
        print(start_coord)
        side_chain_trace = self.trace_side_chain(start_coord, c_coord, n_coord,
                                                 prev_ca_coord, next_ca_coord,
                                                 self.interpolator, 1.6)

        if len(side_chain_trace) > 0:
            ang = self.path_bb_angle(side_chain_trace, c_coord, n_coord)
            # print(ang)
            if ang < math.pi / 4:
                if cb_coord != []:
                    side_chain_trace = [cb_coord]
                else:
                    side_chain_trace = []




        # print('Trace', time.time() - start)

        #if len(sidechain_trace) > 1:
        #    sidechain_trace = self.exclude_gaps(sidechain_trace)
        residue_trace = [c_coord, ca_coord] + side_chain_trace
        # print(residue_trace)
        residue_trace = self.connect_trace(residue_trace)

        trace_indices = self._coord_lst_to_indices_lst_int(residue_trace, self.cube_map)
        trace_map = self._points_to_map(trace_indices, self.cube_map, 3)

        # print(self.operations_counter)
        start = time.time()
        mask = self.mask_trace(self.cube_map, trace_indices, 1.5, 1)
        # print('Mask', time.time() - start)

        indices = self.coord_to_index_int(c_coord, self.cube_map)
        self.walk_grid[indices] = 1.7
        indices = self.coord_to_index_int(n_coord, self.cube_map)
        self.walk_grid[indices] = 1.7
        indices = self.coord_to_index_int(ca_coord, self.cube_map)
        self.walk_grid[indices] = 1.7

        walk_map = MapParser('')
        walk_map.copy_header(self.cube_map)
        walk_map.data = self.walk_grid

        residue_map = MapParser('')
        residue_map.copy_header(self.cube_map)
        residue_map.data = self.cube_map.data * mask

        return self.cube_map, trace_map, walk_map, residue_map

    @staticmethod
    def unit_vector(v):
        return v / np.linalg.norm(v)

    def path_bb_angle(self, path, c_coord, n_coord):
        aver = sum(path) / len(path)
        v1 = c_coord - n_coord
        v2 = c_coord - aver
        v3 = n_coord - aver

        v1 = self.unit_vector(v1)
        v2 = self.unit_vector(v2)
        v3 = self.unit_vector(v3)

        ang1 = np.arccos(np.clip(np.dot(v1, v2), -1.0, 1.0))
        if ang1 > math.pi / 2:
            ang1 = math.pi - ang1
        ang2 = np.arccos(np.clip(np.dot(v1, v3), -1.0, 1.0))
        if ang2 > math.pi / 2:
            ang2 = math.pi - ang2
        return min([ang1, ang2])


    @staticmethod
    def _get_bb_coordinates(model):
        """
        Extracts the C, CA, N and CB coordinates.
        :param model: Biopython residue object
        :return: list of coordinates
        """
        ca = []
        cb = []
        c = []
        n = []
        for atom in model.get_atoms():
            if atom.get_name() == 'CA':
                ca = np.array(atom.coord)
            if atom.get_name() == 'C':
                c = np.array(atom.coord)
            if atom.get_name() == 'N':
                n = np.array(atom.coord)
            if atom.get_name() == 'CB':
                cb = np.array(atom.coord)
        return c, ca, n, cb

    def _coord_lst_to_indices_lst_int(self, coord_lst, map_obj):
        """
        Converts a list of coordinates (in angstroms) to a list of map integer indices
        :param coord_lst: list of coordinates
        :param map_obj: map object
        :return: list of map indices
        """
        indices_lst = []
        for coordinate in coord_lst:
            try:
                indices = self.coord_to_index_int(coordinate, map_obj)
            except IndexError:
                pass
            else:
                indices_lst.append(indices)
        return indices_lst

    @staticmethod
    def _points_to_map(indices_lst, map_template, value):
        """
        Creates a map object of the size of the 'map_template'. Sets the
        map values at 'indices_lst' positions to 'value', all other map values are 0.
        :param indices_lst: list of map indices
        :param map_template: map object
        :return: map object
        """
        out_map = MapParser('')
        grid = np.zeros(map_template.data.shape, dtype='float32')
        i = 0.1
        for indices in indices_lst:
            grid[indices] = value - i
            i += 0.1
        out_map.copy_header(map_template)
        out_map.data = grid
        return out_map

    def mask_trace(self, map_obj, trace, radius, soft_radius):
        mask = np.zeros(map_obj.data.shape, dtype='float32')
        delta2 = radius + soft_radius
        delta1 = radius
        voxel_size = map_obj.voxel_size[0]
        for point in trace:

            rx = int(round(delta2 / voxel_size))
            ry = int(round(delta2 / voxel_size))
            rz = int(round(delta2 / voxel_size))
            for x in range(point[0] - rx, point[0] + rx):
                for y in range(point[1] - ry, point[1] + ry):
                    for z in range(point[2] - rz, point[2] + rz):
                        # Calculate the distance between the current atom and the current voxel
                        d = voxel_size * math.sqrt((x - point[0]) ** 2 + (y - point[1]) ** 2 + (z - point[2]) ** 2)

                        # Assign mask values based to the distance to the atoms
                        try:
                            if d < delta1:
                                mask[x, y, z] = 1
                            elif delta1 < d < delta2:
                                mask[x, y, z] += (math.cos((math.pi / soft_radius) * (d - delta1)) + 1) / 2
                                # if intensity value became > 1 it is set to 1
                                if mask[x, y, z] > 1:
                                    mask[x, y, z] = 1
                        except IndexError:
                            pass
        return mask

    def connect_trace(self, trace, max_dist=0.8):
        """

        :param trace:
        :param max_dist:
        :return:
        """
        connected_trace = trace[:1]
        for index, point in enumerate(trace[1:]):
            point0 = trace[index-1]
            d = self.distance(point0, point)
            if d > max_dist:
                connection_points = int(round(d / max_dist))
                delta = (point - point0) / connection_points
                for i in range(connection_points):
                    new_point = point0 + delta * i
                    connected_trace.append(new_point)
            connected_trace.append(point)
        return connected_trace

    def exclude_gaps(self, trace):
        """

        :param trace:
        :return:
        """
        connected_trace = [trace[0]]
        for index, point in enumerate(trace[1:]):
            point0 = trace[index]
            d = self.distance(point0, point)
            if d > 2:
                deep_min = self.check_minimum(point0, point)
                if deep_min:
                    break
                else:
                    connected_trace.append(point)
            else:
                connected_trace.append(point)
        return connected_trace

    def connect_trace_(self, trace, max_dist=0.8):
        # print(trace)
        connected_trace = [trace[0]]
        for index, point in enumerate(trace[1:]):
            point0 = trace[index-1]
            d = self.distance(point0, point)
            # print('D', d)
            if d > max_dist:
                deep_min = self.check_minimum(point0, point)
                if deep_min:
                    # print('DEEP MIN')
                    break
                connection_points = int(round(d / max_dist))
                delta = (point - point0) / connection_points
                for i in range(connection_points):
                    new_point = point0 + delta * i
                    connected_trace.append(new_point)
                connected_trace.append(point)
            else:
                connected_trace.append(point)
        return connected_trace

    def check_minimum(self, p1, p2, depth=0.2):
        """
        Detects regions where the density of different residues might be connected.
        First takes two points in a map, connects them with a line and calculates the map values along the line. Then
        interpolates the map values using cubic spline and finds the critical points and values on the p1-p2 segment.
        For each minimum checks the flanking maximums and if the average maximum is more than 'depth' higher than
        the minimum returns True else False.
        :param p1: point (np array)
        :param p2: point (np array)
        :param depth: min max relative difference
        :return: True or False
        """
        deep_min = False
        interim_values = []
        v = p2 - p1
        for i in range(20):
            coord = p1 + i * v/20
            indices = self.coord_to_index(coord, self.cube_map)
            val = self.interpolator(indices)
            interim_values.append(val)
        cr_pts, cr_vals = self.critical_points([i for i in range(20)], interim_values)
        min_index = np.argmin(cr_vals)

        if min_index < len(cr_pts) - 2:
            max1 = -1
            max1_ind = None
            max2 = 10000
            max2_ind = None
            for ind, point in enumerate(cr_pts):
                if max1 < point < cr_pts[min_index]:
                    max1 = point
                    max1_ind = ind
                elif cr_pts[min_index] < point < max2:
                    max2 = point
                    max2_ind = ind
            av_max = (cr_vals[max1_ind] + cr_vals[max2_ind]) / 2
            gap = (av_max - cr_vals[min_index]) / av_max
            if gap > depth:
                deep_min = True
        return deep_min

    @staticmethod
    def critical_points(x_axis, y_axis):
        """
        Finds the critical points of a 1D interpolated function
        :param x_axis: x-values, array-like
        :param y_axis: y-values, array-like
        :return: critical_points list, critical_values list
        """
        def quadratic_spline_roots(spl):
            roots = []
            knots = spl.get_knots()
            for a, b in zip(knots[:-1], knots[1:]):
                u, v, w = spl(a), spl((a + b) / 2), spl(b)
                t = np.roots([u + w - 2 * v, w - u, 2 * v])
                t = t[np.isreal(t) & (np.abs(t) <= 1)]
                roots.extend(t * (b - a) / 2 + (b + a) / 2)
            return np.array(roots)

        f = InterpolatedUnivariateSpline(x_axis, y_axis, k=3)
        cr_pts = quadratic_spline_roots(f.derivative())
        cr_pts = np.append(cr_pts, (x_axis[0], x_axis[-1]))
        cr_vals = f(cr_pts)
        return cr_pts, cr_vals

    def find_local_max_map_value(self, index, map_obj, radius):
        # print(map_obj.data[index])
        values = []
        for i in range(index[0]-radius, index[0]+radius):
            for j in range(index[1] - radius, index[1] + radius):
                for k in range(index[2] - radius, index[2] + radius):
                    val = map_obj.data[i, j, k]
                    values.append(val)
        max_val = max(values)
        # print('Mean',max_val)
        return max_val

    def trace_side_chain_(self, start_coord, c_coord, n_coord, prev_ca_coord, next_ca_coord, interpolator,
                         max_radius, thld_step=0.02, thld_delta=0.01):

        path = []
        bb_vector = n_coord - c_coord
        ort_vector = self.orthogonal_vector(bb_vector)
        bb_axis = bb_vector / LA.norm(bb_vector)

        thld = self.max_thld * 0.9
        min_thld = self.max_thld * 0.3

        ca_int = self.interpolator(self.coord_to_index(start_coord, self.cube_map))
        # print('Start I', ca_int)
        start_index = self.coord_to_index_int(start_coord, self.cube_map)
        loc_max = self.find_local_max_map_value(start_index, self.cube_map, 1)


        # print('THLD', thld)
        # print('Min thld', min_thld)

        fused = True
        while fused:
            fused = False
            path = []
            d = 0
            search_center = start_coord
            #thld = self.max_thld * 0.9
            thld = loc_max * 0.9
            radius = 0.4
            while thld > min_thld and d < 10:
                delta = thld_delta * self.max_thld
                thld_range = [thld - delta, thld + delta]
                # print('THLD range', thld_range)
                found = False

                local_max_d = 0
                thld_center = search_center
                while not found:

                    max_dist_coord, d = self._furthest_from_center(c_coord, bb_axis,
                                                                   ort_vector, search_center,
                                                                   prev_ca_coord, next_ca_coord,
                                                                   interpolator, thld_range, radius)
                    # print("d!!", d)

                    if d == 0 and radius < max_radius:
                        radius += 0.1
                    elif d > local_max_d:
                        search_center = max_dist_coord
                        local_max_d = d
                    else:
                        found = True
                        if max_dist_coord != []:
                            path.append(max_dist_coord)

                rolling_dist = self.distance(thld_center, search_center)
                if rolling_dist > self.max_roll:
                    fused = True
                    min_thld = min_thld + thld_step * self.max_thld
                    break

                thld = thld - self.max_thld * thld_step

        return path


    def trace_side_chain(self, start_coord, c_coord, n_coord, prev_ca_coord, next_ca_coord, interpolator,
                         max_radius, thld_step=0.02, thld_delta=0.01):

        path = []
        bb_vector = n_coord - c_coord
        ort_vector = self.orthogonal_vector(bb_vector)
        bb_axis = bb_vector / LA.norm(bb_vector)

        thld = self.max_thld * 0.9
        min_thld = self.max_thld * 0.3

        ca_int = self.interpolator(self.coord_to_index(start_coord, self.cube_map))
        # print('Start I', ca_int)
        start_index = self.coord_to_index_int(start_coord, self.cube_map)
        loc_max = self.find_local_max_map_value(start_index, self.cube_map, 1)


        # print('THLD', thld)
        # print('Min thld', min_thld)

        fused = True
        while fused:
            fused = False
            path = []
            d = 0
            search_center = start_coord
            #thld = self.max_thld * 0.9
            thld = loc_max * 0.9
            radius = 0.4
            while thld > min_thld and d < 10:
                delta = thld_delta * self.max_thld
                thld_range = [thld - delta, thld + delta]
                # print('THLD range', thld_range)
                found = False

                local_max_d = 0
                thld_center = search_center
                while not found:

                    max_dist_coord, d = self._furthest_from_center(c_coord, bb_axis,
                                                                   ort_vector, search_center,
                                                                   prev_ca_coord, next_ca_coord,
                                                                   interpolator, thld_range, radius)
                    # print("d!!", d)

                    # This stops the search if it goes too close to the flanking CA
                    if prev_ca_coord is not None and d:
                        prev_ca_d = self.distance(max_dist_coord, prev_ca_coord)
                    else:
                        prev_ca_d = 3
                    if next_ca_coord is not None and d:
                        next_ca_d = self.distance(max_dist_coord, next_ca_coord)
                    else:
                        next_ca_d = 3
                    if prev_ca_d < 3 and next_ca_d < 3:
                        return path

                    if d == 0 and radius < max_radius:
                        radius += 0.1
                    elif d > local_max_d:
                        search_center = max_dist_coord
                        local_max_d = d
                    else:
                        found = True
                        if max_dist_coord != []:
                            path.append(max_dist_coord)

                rolling_dist = self.distance(thld_center, search_center)
                if rolling_dist > self.max_roll:
                    fused = True
                    min_thld = min_thld + thld_step * self.max_thld
                    break

                thld = thld - self.max_thld * thld_step

        return path


    def _furthest_from_center(self, bb_point, bb_axis, ort_vector, search_center, prev_ca_coord, next_ca_coord,
                              interpolator, thld_range, radius):

        max_dist_coord = []
        max_dist = 0

        x0 = search_center + ort_vector * radius
        x0 = x0 - bb_axis * 0.2 * 4

        for i in range(9):
            x = x0 + bb_axis * 0.2 * i
            for j in range(16):
                self.operations_counter += 1
                alpha = math.pi / 8 * j

                matrix = self.rotation_matrix(bb_axis, alpha)
                x1 = np.dot(matrix, x)
                y1 = np.dot(matrix, search_center)
                t = y1 - search_center
                x1 = x1 - t

                int_indices = self.coord_to_index_int(x1, self.cube_map)
                try:
                    if self.walk_grid[int_indices] == 0:
                        self.walk_grid[int_indices] = 1
                except IndexError:
                    pass

                indices = self.coord_to_index(x1, self.cube_map)
                try:
                    map_value = interpolator(indices)
                except ValueError:
                    continue

                if thld_range[0] < map_value < thld_range[1]:
                    d = self.distance_to_line(x1, bb_point, vector=bb_axis)


                    if d > max_dist:
                        max_dist_coord = x1
                        max_dist = d
                        if self.walk_grid[int_indices] == 0:
                            self.walk_grid[int_indices] = 1.2
            center_indices = self.coord_to_index_int(search_center, self.cube_map)
            self.walk_grid[center_indices] = 1.5
        # print(self.operations_counter)
        return max_dist_coord, max_dist

    @staticmethod
    def distance(p1, p2):
        """
        Calculates the distance between 2 points in 3D
        :param p1: point coordinates
        :param p2: point coordinates
        :return: distance
        """
        # print(p1, p2)
        d = math.sqrt((p1[0] - p2[0]) ** 2
                      + (p1[1] - p2[1]) ** 2
                      + (p1[2] - p2[2]) ** 2)
        return d

    @staticmethod
    def distance_to_line(x0, x1, x2=None, vector=None):
        """
        Calculates the distance between a point and a line. The line can be defined by 2 points or a point and a vector.
        :param x0: point coordinates (np.array)
        :param x1: point (np.array)
        :param x2: point (np.array)
        :param vector: vector (np.array)
        :return: distance
        """
        if vector is None and x2 is not None:
            vector = x2 - x1
        elif x2 is None and vector is None:
            raise Exception("A direction vector and a point or two points are required to define the line")
        d = np.linalg.norm(np.cross(vector, (x1-x0))) / np.linalg.norm(vector)
        return d

    @staticmethod
    def distance_to_line_points(x0, x1, x2):

        d = math.fabs(np.cross(x0-x1, x0-x2)) / math.fabs(x2 - x1)
        return d

    @staticmethod
    def orthogonal_vector_at_point(point, vector):
        """delete"""
        if vector[2] != 0:
            z = (vector[0]*point[0] + vector[1]*point[1]) / vector[2]
            ort_vector = np.array((round(point[0], 2), round(point[1], 2), round(z, 2)))
        elif vector[1] != 0:
            y = (vector[0] * point[0] + vector[2] * point[2]) / vector[1]
            ort_vector = np.array((round(point[0], 2), round(y, 2), round(point[2], 2)))
        else:
            x = (vector[1] * point[1] + vector[2] * point[2]) / vector[0]
            ort_vector = np.array((round(x, 2), round(point[1], 2), round(point[2], 2)))

        return ort_vector

    @staticmethod
    def orthogonal_vector(v):
        """
        Finds a vector orthogonal to the input vector
        :param v: vector
        :return: unit orthogonal vector
        """
        if v[0] != 0:
            y = randint(1, 2)
            z = randint(1, 2)
            x = (v[1] * y + v[2] * z) / -v[0]
            ort_vector = np.array((x, y, z))

        elif v[1] != 0:
            x = randint(1, 2)
            z = randint(1, 2)
            y = (v[2] * z + v[0] * x) / -v[1]
            ort_vector = np.array((x, y, z))
        else:
            x = randint(1, 2)
            y = randint(1, 2)
            z = (v[0] * x + v[1] * y) / -v[2]
            ort_vector = np.array((x, y, z))
        ort_vector = ort_vector / LA.norm(ort_vector)
        return ort_vector

    @staticmethod
    def line_function(point, vector):
        """
        Defines a function F(x) of a line in 3D which
        :param point: a point on the line
        :param vector: line direction vector
        :return: line function
        """
        def f(x):
            y = (x - point[0]) * vector[1] / vector[0] + point[1]
            z = (x - point[0]) * vector[2] / vector[0] + point[2]
            coord = (x, y, z)
            return coord

        return f

    @staticmethod
    def coord_to_index_int(coord, map_obj):
        """
        Converts atomic residue coordinates into integer map indices
        :param coord: coordinate in angstrom
        :param map_obj: map object
        :return: int indices
        """
        map_obj.origin = np.array(map_obj.origin)
        st = map_obj.n_start
        vs = map_obj.voxel_size
        orig = map_obj.origin
        x_index = int(round(coord[2] / vs[2] - orig[2] / vs[2] - st[2]))
        y_index = int(round(coord[1] / vs[1] - orig[1] / vs[1] - st[1]))
        z_index = int(round(coord[0] / vs[0] - orig[0] / vs[0] - st[0]))
        return x_index, y_index, z_index

    @staticmethod
    def coord_to_index(coord, map_obj):
        """
        Converts atomic residue coordinates into map indices
        :param coord: coordinate in angstrom
        :param map_obj: map object
        :return: float indices
        """
        st = map_obj.n_start
        vs = map_obj.voxel_size
        orig = map_obj.origin
        x_index = coord[2] / vs[2] - orig[2] / vs[2] - st[2]
        y_index = coord[1] / vs[1] - orig[1] / vs[1] - st[1]
        z_index = coord[0] / vs[0] - orig[0] / vs[0] - st[0]
        return x_index, y_index, z_index

    @staticmethod
    def rotation_matrix(axis, theta):
        """
        Return the rotation matrix associated with counterclockwise rotation about
        the given axis by theta radians.
        """
        axis = np.asarray(axis)
        axis = axis / math.sqrt(np.dot(axis, axis))
        a = math.cos(theta / 2.0)
        b, c, d = -axis * math.sin(theta / 2.0)
        aa, bb, cc, dd = a * a, b * b, c * c, d * d
        bc, ad, ac, ab, bd, cd = b * c, a * d, a * c, a * b, b * d, c * d
        return np.array([[aa + bb - cc - dd, 2 * (bc + ad), 2 * (bd - ac)],
                         [2 * (bc - ad), aa + cc - bb - dd, 2 * (cd + ab)],
                         [2 * (bd + ac), 2 * (cd - ab), aa + dd - bb - cc]])

    @staticmethod
    def get_threshold(map_data, level):
        out_map = np.copy(map_data)
        it = np.nditer(out_map, flags=['multi_index'], op_flags=['writeonly'])
        while not it.finished:
            if it[0] < level:
                it[0] = 0
            it.iternext()
        return out_map

    def trace_length(self, trace):
        length = 0
        if len(trace) > 2:
            a1 = trace[0]
            for a2 in trace:
                d = self.distance(a1, a2)
                length += d
                a1 = a2
        return length




from Bio.PDB.PDBParser import PDBParser
from Bio.PDB.MMCIFParser import MMCIFParser, FastMMCIFParser
from Bio.PDB.StructureBuilder import StructureBuilder
from Bio.PDB.PDBIO import PDBIO
from Bio.PDB.MMCIF2Dict import MMCIF2Dict
from Bio.PDB.mmcifio import MMCIFIO
from Bio.PDB.StructureBuilder import StructureBuilder
from Bio.PDB.Polypeptide import PPBuilder
import sys
import string
import json
import copy
import os
import argparse
import subprocess
import numpy as np

AA_3LET_TO_1LET = {'ALA': 'A',
                   'ARG': 'R',
                   'ASN': 'N',
                   'ASP': 'D',
                   'CYS': 'C',
                   'GLU': 'E',
                   'GLN': 'Q',
                   'GLY': 'G',
                   'HIS': 'H',
                   'ILE': 'I',
                   'LEU': 'L',
                   'LYS': 'K',
                   'MET': 'M',
                   'PHE': 'F',
                   'PRO': 'P',
                   'SER': 'S',
                   'THR': 'T',
                   'TRP': 'W',
                   'TYR': 'Y',
                   'VAL': 'V',
                   'HIC': 'H'}


def load_structure(model_path):
    """
    Creates Biopython structure object from file
    :param model_path: pdb, cif file in_dir
    :return: Biopython structure object
    """
    def copy_label_seq_id(in_path):
        parser = MMCIFParser(QUIET=True)
        parser._mmcif_dict = MMCIF2Dict(in_path)
        if "_atom_site.auth_seq_id" in parser._mmcif_dict:
            for i in parser._mmcif_dict["_atom_site.auth_seq_id"]:
                try:
                    int(i)
                except ValueError:
                    parser._mmcif_dict["_atom_site.auth_seq_id"] = parser._mmcif_dict["_atom_site.label_seq_id"]
        parser._build_structure(in_path.split('/')[-1].split('.')[0])
        return parser._structure_builder.get_structure()

    if model_path.split('.')[-1] == 'pdb' or model_path.split('.')[-1] == 'ent':
        parser = PDBParser(PERMISSIVE=1, QUIET=True)
        structure = parser.get_structure(model_path.split('/')[-1].split('.')[0], model_path)
    elif model_path.split('.')[-1] == 'cif':
        try:
            parser = MMCIFParser(QUIET=True)
            structure = parser.get_structure(model_path.split('/')[-1].split('.')[0], model_path)
        except ValueError:
            structure = copy_label_seq_id(model_path)
    else:
        raise Exception('Please provide the input residue in pdb or cif format')

    return structure

def save_model(model, out_path, preserve_atom_numbering=False):
    """
    Save residue to file
    :param model: structure object
    :param out_path: output file in_dir
    """
    ext = out_path.split('.')[-1]
    if ext == 'pdb' or ext == 'ent':
        io = PDBIO()
    else:
        io = MMCIFIO()
    io.set_structure(model)
    io.save(out_path, preserve_atom_numbering=preserve_atom_numbering)

def calc_residue_static_pairwise_rmsd(model1, model2):
    """
    Calculate pairwise RMSD for residue structure objects without superimposing
    Takes into account residues symmetry
    :param model1: structure object
    :param model2: structure object
    :return: RMSD value
    """
    atoms1 = [atom for atom in model1.get_atoms() if atom.get_name().upper() in ['C', 'CA', 'N', 'O']]
    atoms2 = [atom for atom in model2.get_atoms() if atom.get_name().upper() in ['C', 'CA', 'N', 'O']]
    # print(atoms1)
    # print(atoms2)
    if len(atoms2) != len(atoms1):
        return 100
    atoms1.sort(key=lambda x: x.get_name())
    atoms2.sort(key=lambda x: x.get_name())

    array1, array2 = np.array([atoms1[0].get_coord()]), np.array([atoms2[0].get_coord()])
    for i in range(1, len(atoms1)):
        array1 = np.concatenate((array1, np.array([atoms1[i].get_coord()])), axis=0)
        array2 = np.concatenate((array2, np.array([atoms2[i].get_coord()])), axis=0)

    return rmsd(array1, array2)

def rmsd(array1, array2):
    """
    Calculates RMSD between two sets of coordinates without alignment
    :param array1: numpy 2d array
    :param array2: numpy 2d array
    :return: rmsd
    """
    if np.shape(array1) != np.shape(array1):
        raise Exception('The coordinate arrays must have the same dimensions')
    dif = array1 - array2
    return np.sqrt(np.mean(np.sum(dif * dif, axis=1)))


def filter_far_residues(target, reference, max_rms):
    del_r = []
    for t_res in target.get_residues():
        tmp = []
        r_tmp = []
        for r_res in reference.get_residues():
            # print(t_res, r_res)
            tmp.append(calc_residue_static_pairwise_rmsd(t_res, r_res))
            r_tmp.append(r_res)
        min_r = min(tmp)
        index = tmp.index(min(tmp))
        r = r_tmp[index]
        print(t_res.resname, t_res.parent.id, t_res.id,'--', r.resname, r.parent.id, r.id, min(tmp))
        if min_r > max_rms:
            del_r.append(t_res)
        else:
            t_res.resname = r.resname

    out = copy.deepcopy(target)[0]
    for r in del_r:
        ch = r.get_parent().get_id()
        del out[ch][r.get_id()]
    return out

def del_sidechain(model):
    del_atoms = []
    bb = ['CA', 'C', 'N', 'O', 'CB']
    for chain in model:
        for residue in chain:
            for atom in residue:
                if atom.get_name().upper() not in bb:
                    del_atoms.append((chain.get_id(), residue.get_id(), atom.get_id()))
    for chain, resi, atom in del_atoms:
        del model[chain][resi][atom]
    return model


def match_backbones(target_mod, ref_mod, max_rms):
    target = load_structure(target_mod)
    reference = load_structure(ref_mod)

    out = filter_far_residues(target, reference, max_rms)
    return out
from itertools import permutations
def filter_chains(model, min_len, max_bond):

    # for res in model.get_residues():
    #     res.parent.id = '99'

    ids = string.ascii_uppercase + string.ascii_lowercase
    # ids = ids[::-1]
    short_ids = []

    sb = StructureBuilder()
    sb.init_structure('structure')
    sb.init_seg(' ')
    # res_model_id = {0}


    ch_id = 0
    residues = [r for r in model.get_residues()]
    sb.init_model(0)
    sb.init_chain(ids[0])
    sb.structure[0][ids[0]].add(residues[0].copy())

    for i, res in enumerate(residues[:-1]):
        # a = [c.id for c in model.get_chains()]
        # print(a)
        n = 0
        c = 0
        next_r = residues[i+1]
        for atom in res:
            if atom.get_name() == 'C':
                c = atom
        for atom in next_r:
            if atom.get_name() == 'N':
                n = atom
        d = n - c
        if d > max_bond:
            print(ids[ch_id])
            print(d)
            ch_id += 1
            if ch_id > len(ids)-1:
                ids = [''.join(i) for i in permutations(ids, 2)]
                print(ids)
            sb.init_chain(ids[ch_id])
        sb.structure[0][ids[ch_id]].add(next_r.copy())
    short = [c.id for c in sb.structure.get_chains() if len(c) < min_len]
    for c in short:
        del sb.structure[0][c]

    # for i, c in enumerate(model.get_chains()):
    #     c.id = ids[i]
    return sb.structure


# t = '/Volumes/data/filter/build1.pdb'
# r = '/Volumes/data/filter/6272_shifted.pdb'
# o = '/Volumes/data/filter/build1_filt15_ala_rename.pdb'
# t = '/Users/andrei/filter_build/20584/build1.pdb'
# r = '/Users/andrei/filter_build/20584/6tys.pdb'
# o = '/Users/andrei/filter_build/20584/built_filt.pdb'

t = '/Volumes/data/filter/0703/build1.pdb'
r = '/Volumes/data/filter/0703/6kku.cif'
o = '/Volumes/data/filter/0703/0703_built_filt1.cif'
o_tmp = '/Volumes/data/filter/0703/tmp_0703_built_filt.cif'
o_pdb = '/Volumes/data/filter/0703/tmp_0703_built_filt.pdb'
out = match_backbones(t, r, 1.5)
out = del_sidechain(out)
save_model(out, o_tmp)
out = load_structure(o_tmp)[0]
out = filter_chains(out, 10, 1.5)
save_model(out, o)
save_model(out, o_pdb)


"""
penultimateClassifier.py

Groups residues of the same type into Richardson's rotamer classes.
This code uses the Penultimate Rotamer library 'SC Lovell, JM Word,
JS Richardson and DC Richardson (2000) "The Penultimate Rotamer Library"
Proteins: Structure Function and Genetics 40: 389-408.'

Copyright [2013] EMBL - European Bioinformatics Institute
Licensed under the Apache License, Version 2.0 (the
"License"); you may not use this file except in
compliance with the License. You may obtain a copy of
the License at
http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied. See the License for the
specific language governing permissions and limitations
under the License.
"""

__author__ = 'Andrei Istrate'
__email__ = 'andrei@ebi.ac.uk'
__date__ = '2018-05-29'


import sys
import traceback
import math
import os
import copy
import argparse
import json
import time
import numpy as np
from datetime import datetime
from Bio.PDB import calc_dihedral
from Bio.PDB.Superimposer import Superimposer

import threed_strudel.utils.functions as func
import threed_strudel.nomenclature as nomenclature
from threed_strudel.utils import bioUtils
import threed_strudel.lib.strudel.penultimate_rotamer_lib as p_lib


class PenultimateClassifier:
    """
    Class for rotamer classification using a rotamer library in json format
    By default penultimate library (inp/penultimate_lib.json)
    Lovell, S.C., Word, J.M., Richardson, J.S. & Richardson, D.C. "The penultimate rotamer library",
    Proteins: Structure, Function and Genetics, Vol.40, 389-408 (2000)
    """

    def __init__(self, in_dir,
                 out_dir=None,
                 out_prefix='rot_',
                 default_angle_width=30,
                 use_minimal_width=False,
                 warning_level='info',
                 residue_type=None,
                 rotamer_lib_path=None,
                 log_path=None):
        """
        :param in_dir: Input directory containing residues of the same type as separate files
        :param out_dir: Output directory
        :param out_prefix: prefix for output rotamer folders
        :param rotamer_lib_path: penultimate library file (json format)
        :param default_angle_width: rotamer angle width to be used when missing in the library or
               when minimal width flag is True
        :param use_minimal_width: BOOLEAN
        :param warning_level: info, debug
        :param log_path: path for the log
        """
        self.in_dir = in_dir
        self.log_path = log_path
        self.corrupted_files_dir = os.path.join(in_dir, 'corrupted_files')
        if not os.path.exists(self.corrupted_files_dir):
            os.mkdir(self.corrupted_files_dir)
        self.model_err_out_path = self.setup_model_read_error_out()
        self.model_err_out = open(self.model_err_out_path, 'w')

        if out_dir is None:
            self.out_dir = in_dir
        else:
            self.out_dir = out_dir
        self.log = self.setup_log(warning_level)

        self.rotamer_lib = self.load_rotamer_lib(rotamer_lib_path)
        self.out_prefix = out_prefix
        # Set True to use the default width as minimal width value
        self.use_minimal_width = use_minimal_width
        # To be used when the width is not defined in the library and when use_minimal_width=True
        self.default_width = default_angle_width
        if residue_type is not None:
            self.residue_type = residue_type.lower()
        else:
            self.residue_type = residue_type
        self.rotamer_classes = None
        self.rotamers = None

        self.all_res_symm_pairs = nomenclature.SYMMETRIC_ATOM_PAIRS
        self.symmetric_chi = nomenclature.SYMMETRIC_CHI
        self.rotamers_data_structure = self.set_rotamers_data_structure()
        # self.run_classification()

    def run_classification(self):
        """

        :return:
        """
        start = time.time()
        date_time = datetime.now().strftime("%H:%M:%S %Y-%m-%d")
        text = '{:_^100}'.format('PenultimateClassifier') + '\n\nStarted: {}'.format(date_time)
        self.log.info('%s\n', text)

        struct_object_list = self.create_model_list()
        if self.residue_type is None:
            self.residue_type = self.get_res_type(struct_object_list[0])

        self.superimpose_all(struct_object_list)
        self.log.info('Classifying %s residues', self.residue_type.upper())

        if self.residue_type.upper() in ('ALA', 'GLY'):
            self.rotamer_classes = [struct_object_list, []]
            self.rotamer_classes[0][0].id = self.rotamer_classes[0][0].id + '_representative'
        else:
            rotamers = self.create_rotamers(self.residue_type)
            self.rotamers = rotamers
            self.rotamer_classes = self.find_rotamers(struct_object_list, rotamers)

            for class_ in self.rotamer_classes:
                if len(class_) > 200:
                    repr_index = self.find_repr_model_fast(class_, rotamers)
                else:
                    repr_index = self.find_repr_model(class_)
                if repr_index is not None:
                    class_[repr_index].id = class_[repr_index].id + '_representative'

        self.save_rotamers(self.rotamer_classes)
        text = func.report_elapsed(start)
        self.log.info(text)

    @staticmethod
    def load_rotamer_lib(rotamer_lib_path=None):
        """
        Reads the rotamer library in json format
        :param rotamer_lib_path: json file
        :return: file content
        """
        if rotamer_lib_path is not None:
            try:
                with open(rotamer_lib_path) as j:
                    data = json.load(j)
                    return data
            except:
                pass
        data = p_lib.PENULTIMATE_2000
        return data

    def setup_log(self, warning_level):
        """
        Setup log
        :param warning_level:
        :return:
        """
        class_log = func.setup_logger('log', self.log_path, warning_level=warning_level)
        return class_log

    def setup_model_read_error_out(self):
        if self.log_path is not None:
            err_out_file = os.path.splitext(self.log_path)[0] + '.err'
        else:
            err_out_file = os.path.join(self.corrupted_files_dir, 'corrupted_files.err')
        return err_out_file

    def create_model_list(self):
        """
        Reads all .cif files in a directory and creates a list of structure objects
        """
        if self.residue_type is not None:
            r_type = self.residue_type
            type_defined = True
        else:
            type_defined = False
            r_type = None

        files = os.listdir(self.in_dir)
        cif_names = [i for i in files if (i.endswith('.cif') or i.endswith('.pdb') or i.endswith('.ent'))
                     and not i.startswith('.')]

        str_object_list = []
        for element in cif_names:
            try:
                structure = bioUtils.load_structure(os.path.join(self.in_dir, element))
            except:
                # If the residue file could not be read for some reason if will move it
                self.log.info('Failed to process %s file\nMoving it to and related files %s\nCheck %s for details',
                              element, self.corrupted_files_dir, self.model_err_out_path)
                exc_type, exc_value, exc_traceback = sys.exc_info()
                traceback.print_tb(exc_traceback, file=self.model_err_out)
                # os.replace(os.path.join(self.in_dir, element), os.path.join(self.corrupted_files_dir, element))
                prefix = element.split('_')[0]
                related_files = [i for i in files if i.startswith(prefix)]
                if related_files:
                    for file in related_files:
                        os.replace(os.path.join(self.in_dir, file), os.path.join(self.corrupted_files_dir, file))
            else:
                tt = self.get_res_type(structure)
                if tt is not None:
                    if r_type is None:
                        r_type = tt
                    if tt == r_type:
                        str_object_list.append(structure)
                    elif not type_defined:
                        raise Exception('There are multiple residue types in the input and no '
                                        '"residue_type" parameter specified. '
                                        '\nPlease restart with defined "residue_type" parameter')

        if not str_object_list:
            raise Exception(f'There are no atomic residue files in the specified directory: {self.in_dir}')
        return str_object_list

    @staticmethod
    def set_rotamers_data_structure():
        """
        A dictionary to define the rotameric angles for standard amino acids.
        :return: dictionary
        """
        residue_rotamers = nomenclature.ROTAMER_DATA
        for key in residue_rotamers.keys():
            residue_rotamers[key]['id'] = None
            residue_rotamers[key]['name'] = None
        return residue_rotamers

    def create_rotamers(self, residue_3let):
        """
        Creates a list of dictionaries with relevant rotameric parameters for the specified residue type
        :param residue_3let: residue code (3 letter)
        :return: list of dictionaries
        """
        residue_3let = residue_3let.lower()
        rotamers = []

        residue_lib_data = self.rotamer_lib[residue_3let]
        for rotamer in residue_lib_data:
            rot_data_structure = copy.deepcopy(self.rotamers_data_structure[residue_3let])
            for key in rot_data_structure.keys():
                if key.startswith('chi_'):
                    rot_data_structure[key]['angle_val'] = rotamer[key]
                    nr = key.split('_')[-1]
                    try:
                        value = int(rotamer['chi-width_' + nr])
                    except TypeError:
                        value = self.default_width

                    if not self.use_minimal_width:
                        rot_data_structure[key]['angle_width'] = value
                    else:
                        rot_data_structure[key]['angle_width'] = max([value, self.default_width])
                else:
                    rot_data_structure[key] = rotamer[key]
            rotamers.append(rot_data_structure)

        return rotamers

    @staticmethod
    def get_res_type(model):
        """
        Reads the 3 letter residue code from a residue  BIO.PDB object
        :param model:
        :return: 3 letter residue code
        """
        residues = [residue for residue in model.get_residues()]
        if len(residues) == 1:
            return residues[0].get_resname().lower()
        else:
            return None

    @staticmethod
    def _get_dihedral_angle(residue, atom_names):

        if len(atom_names) < 4:
            raise Exception(f"Minimum 4 atom names needed, {len(atom_names)} were given")

        v = [None for _ in range(len(atom_names))]
        atoms = residue.get_atoms()
        for atom in atoms:
            for i, name in enumerate(atom_names):
                if atom.get_name().lower() == name:
                    v[i] = atom.get_vector()

        if any([a is None for a in v]):
            missing = []
            for i, v in enumerate(v):
                if v is None:
                    missing.append(atom_names[i])
            raise Exception(f"Missing atoms in the input residue: {' '.join(missing)}")

        angle = calc_dihedral(v[0], v[1], v[2], v[3])
        angle = math.degrees(angle)
        if len(atom_names) > 4:
            angle1 = calc_dihedral(v[0], v[1], v[2], v[4])
            angle1 = math.degrees(angle1)
            angles = [angle, angle1]
        else:
            angles = [angle]
        return angles

    @staticmethod
    def _get_dihedral_angle1(model, atom_names):
        vector1, vector2, vector3, vector4, vector5 = None, None, None, None, None
        atoms = model.get_atoms()
        for atom in atoms:
            if atom.get_name().lower() == atom_names[0]:
                vector1 = atom.get_vector()
            elif atom.get_name().lower() == atom_names[1]:
                vector2 = atom.get_vector()
            elif atom.get_name().lower() == atom_names[2]:
                vector3 = atom.get_vector()
            elif atom.get_name().lower() == atom_names[3]:
                vector4 = atom.get_vector()
            if len(atom_names) == 5:
                if atom.get_name().lower() == atom_names[4]:
                    vector5 = atom.get_vector()

        if any([a is None for a in [vector1, vector2, vector3, vector4]]):
            missing = []
            for i, v in enumerate([vector1, vector2, vector3, vector4]):
                if v is None:
                    missing.append(atom_names[i])
            raise Exception(f"Missing atoms in the input residue: {' '.join(missing)}")

        angle = calc_dihedral(vector1, vector2, vector3, vector4)
        angle = math.degrees(angle)
        if len(atom_names) == 5:
            angle1 = calc_dihedral(vector1, vector2, vector3, vector5)
            angle1 = math.degrees(angle1)
            angles = [angle, angle1]
        else:
            angles = [angle]
        return angles

    def find_rotamers(self, str_object_list, rotamers):
        """
        Group the input models into Richardson classes
        :return: a list of rotamer lists, the last element contains the non-rotameric models
        """
        classes = []
        for _ in rotamers:
            classes.append([])
        classes.append([])

        for model in str_object_list:
            sorted_ = False
            angles = {}
            for key, value in rotamers[0].items():
                if key not in ('name', 'id'):
                    angles[key] = self._get_dihedral_angle(model, value['atoms'])

            for rotamer in rotamers:
                belongs = True
                rotamer_nr = None
                for key, value in rotamer.items():
                    if key == 'id':
                        rotamer_nr = value
                    elif key == 'name':
                        pass
                    else:
                        min_ = value['angle_val'] - value['angle_width'] / 2
                        max_ = value['angle_val'] + value['angle_width'] / 2
                        if len(angles[key]) > 1:
                            inside1 = self.angle_in_interval(angles[key][0], min_, max_)
                            inside2 = self.angle_in_interval(angles[key][1], min_, max_)
                            inside = inside1 or inside2
                        else:
                            inside = self.angle_in_interval(angles[key][0], min_, max_)
                        if not inside:
                            belongs = False
                if belongs and not sorted_:
                    classes[rotamer_nr - 1].append(model)
                    sorted_ = True
            if not sorted_:
                classes[-1].append(model)
        return classes

    @staticmethod
    def angle_in_interval(n, a, b):
        """
        Checks if the angle n belongs to the interval [a, b]
        :param n: angle (deg)
        :param a: angle (deg)
        :param b: angle (deg)
        :return: bool
        """
        n = n % 360
        a = a % 360
        b = b % 360
        if a < b:
            return a <= n <= b
        return a <= n or n <= b

    def save_rotamers(self, classes):
        """
        Save each rotamer class and the non-rotameric models in separate folders
        :param classes: a list of rotamer lists, the last element contains the non-rotameric models
        """
        nr_rotamers = 0
        nr_non_rotamers = 0
        for index, element in enumerate(classes[0:-1]):
            if self.residue_type.upper() in ('ALA', 'GLY'):
                folder = '{}1_{}'.format(self.out_prefix, 'single-conf')
            else:
                folder = '{}{}_{}'.format(self.out_prefix, index + 1, self.rotamers[index]['name'])
            path = os.path.join(self.out_dir, folder)
            if not os.path.exists(path):
                os.makedirs(path)
            for model in classes[index]:
                model_path = os.path.join(path, model.id + '.cif')
                bioUtils.save_model(model, model_path)
                nr_rotamers += 1
            self.concatenate_residue_models(classes[index], os.path.join(path, 'class_' + str(index + 1) + '.cif'))

        path = os.path.join(self.out_dir, 'non-rotameric')
        if not os.path.exists(path):
            os.makedirs(path)
        for model in classes[-1]:
            model_path = os.path.join(path, model.id + '.cif')
            bioUtils.save_model(model, model_path)
            nr_non_rotamers += 1
        self.concatenate_residue_models(classes[-1], os.path.join(path, self.residue_type + '_non-rotameric.cif'))
        self.log.info("%s %s residues were classified as rotameric\n"
                      "%s %s residues were classified as non rotameric",
                      nr_rotamers, self.residue_type.upper(),
                      nr_non_rotamers, self.residue_type.upper())

    @staticmethod
    def superimpose_n_ca_c(str_fixed, str_moving):
        """
        Superimpose 2 models containing single residues using N, CA, C atom selection
        :param str_fixed: BIO.PDB structure object
        :param str_moving: BIO.PDB structure object
        """
        fixed = []
        moving = []
        for atom in [a for a in str_fixed.get_atoms()]:
            if atom.get_id() == 'CA' or atom.get_id() == 'C' or atom.get_id() == 'N':
                fixed.append(atom)
        for atom in [a for a in str_moving.get_atoms()]:
            if atom.get_id() == 'CA' or atom.get_id() == 'C' or atom.get_id() == 'N':
                moving.append(atom)
        sup = Superimposer()
        sup.set_atoms(fixed, moving)
        sup.apply([a for a in str_moving.get_atoms()])
        return str_moving

    def superimpose_all(self, str_object_list):
        """
        Superimpose all structures to the first one in the list
        """
        i = 1
        while i < len(str_object_list):
            str_object_list[i] = self.superimpose_n_ca_c(str_object_list[0], str_object_list[i])
            i += 1

    @staticmethod
    def concatenate_residue_models(model_obj_list, out_file):
        """
        Merge a list of structure objects files into a multimodel structure object
        :param model_obj_list: residue directory
        :param out_file: output file name
        """
        if len(model_obj_list) == 1:
            bioUtils.save_model(model_obj_list[0], out_file)
        elif len(model_obj_list) > 1:
            out_structure = copy.deepcopy(model_obj_list[0])
            out_structure[0].serial_num = 1
            for i, structure in enumerate(model_obj_list[1:]):
                structure[0].id = i + 1
                out_structure.add(structure[i + 1])
                out_structure[i + 1].serial_num = i + 2
            for model in out_structure:
                for chain in model:
                    chain.id = 'A'
            bioUtils.save_model(out_structure, out_file)

    def set_default_rotamer_angle_width(self, width):
        """
        Set the default dihedral angle width for rotamer parameters setting
        :param width: dihedral angle width
        """
        self.default_width = width

    def find_repr_model(self, model_list):
        """
        Find the representative models of the rotamer classes
        :param model_list: list of residue objects
        """
        self.log.debug("Searching for the representative residue in each rotamer class")
        if len(model_list) == 0:
            return None
        elif len(model_list) < 3:
            return 0
        rmsd_list = []
        for model1 in model_list:
            rmsd = 0
            for model2 in model_list:
                if model1 != model2:
                    rmsd += self.calc_residue_static_pairwise_rmsd(model1, model2)
            rmsd_list.append(rmsd)
        min_rms = min(rmsd_list)

        return rmsd_list.index(min_rms)

    def find_repr_model_fast(self, model_list, rotamers):
        """
        Find the representative models of the rotamer classes.
        The search is done in two steps:
        First based on the distance between the rotameric angles and the averaged rotameric angles.
        Second, 10% best candidates are evaluated using RMSD
        :param model_list: List of residue objects
        :param rotamers: rotamers dictionary
        :return representative residue index
        """
        self.log.debug("Searching for the representative residue in each rotamer class")
        if len(model_list) == 0:
            return None
        # First stage (find candidates)
        all_angles = {}
        for model in model_list:
            for key, value in rotamers[0].items():
                if key not in ('name', 'id'):
                    angles = self._get_dihedral_angle(model, value['atoms'])
                    angle = min([a % 360 for a in angles])
                    if key in all_angles.keys():
                        all_angles[key].append(angle)
                    else:
                        all_angles[key] = [angle]

        average_angles = {}
        for key in all_angles.keys():
            average_angles[key] = sum(all_angles[key]) / len(all_angles[key])

        keys = all_angles.keys()
        weights = []
        for i in range(len(model_list)):
            weight = 0
            for key in keys:
                weight += math.fabs(average_angles[key] - all_angles[key][i])
            weights.append(weight)
        min_indices = []
        max_weight = max(weights)
        if len(weights) > 10:
            ln = int(len(weights)*0.1)
        else:
            ln = len(weights)
        for i in range(ln):
            index = weights.index(min(weights))
            min_indices.append(index)
            weights[index] = max_weight + 1
        # Second stage accurately evaluate the candidates and find the best one
        rmsds = []
        for i in min_indices:
            ref_model = model_list[i]
            rmsd = 0
            for model in model_list:
                rmsd += self.calc_residue_static_pairwise_rmsd(model, ref_model)
            rmsds.append(rmsd)

        return min_indices[rmsds.index(min(rmsds))]

    def calc_residue_static_pairwise_rmsd(self, model1, model2):
        """
        Calculate pairwise RMSD for residue structure objects without superimposing
        Takes into account residues symmetry
        :param model1: structure object
        :param model2: structure object
        :return: RMSD value
        """
        atoms1 = [atom for atom in model1.get_atoms() if not atom.get_name().upper().startswith('H')]
        atoms2 = [atom for atom in model2.get_atoms() if not atom.get_name().upper().startswith('H')]
        if len(atoms1) != len(atoms2):
            self.log.info("Model %s (%s) and %s (%s) have different number of hevy atoms!!!\n"
                          "The calculated rms might be not accurate",
                          model1.id, len(atoms1), model2.id, len(atoms2))
        names1 = set([atom.get_name() for atom in atoms1])
        names2 = set([atom.get_name() for atom in atoms2])
        names = names1.intersection(names2)
        atoms1 = [atom for atom in atoms1 if atom.get_name() in names]
        atoms2 = [atom for atom in atoms2 if atom.get_name() in names]

        atoms1.sort(key=lambda x: x.get_name())
        atoms2.sort(key=lambda x: x.get_name())

        array1, array2 = np.array([atoms1[0].get_coord()]), np.array([atoms2[0].get_coord()])
        for i in range(1, len(atoms1)):
            array1 = np.concatenate((array1, np.array([atoms1[i].get_coord()])), axis=0)
            array2 = np.concatenate((array2, np.array([atoms2[i].get_coord()])), axis=0)
        rmsd = self.rmsd(array1, array2)

        res_type = self.get_res_type(model1).lower()

        if res_type not in self.all_res_symm_pairs.keys():
            return rmsd

        symmetric_pairs = self.all_res_symm_pairs[res_type]
        swap_indices = [[],[]]
        for i in range(0, len(symmetric_pairs), 2):
            for k, atom in enumerate(atoms2):
                if atom.get_name().lower() == symmetric_pairs[i]:
                    swap_indices[0].append(k)
                elif atom.get_name().lower() == symmetric_pairs[i+1]:
                    swap_indices[1].append(k)
        for i in range(len(swap_indices[0])):
            atoms2[swap_indices[0][i]], atoms2[swap_indices[1][i]] = \
                atoms2[swap_indices[1][i]], atoms2[swap_indices[0][i]]

        array3 = np.array([atoms2[0].get_coord()])
        for i in range(1, len(atoms2)):
            array3 = np.concatenate((array3, np.array([atoms2[i].get_coord()])), axis=0)
        rmsd2 = self.rmsd(array1, array3)
        return min(rmsd, rmsd2)

    @staticmethod
    def rmsd(array1, array2):
        """
        Calculates RMSD between two sets of coordinates without alignment
        :param array1: numpy 2d array
        :param array2: numpy 2d array
        :return: rmsd
        """
        if np.shape(array1) != np.shape(array1):
            raise Exception('The coordinate arrays must have the same dimensions')
        dif = array1 - array2
        return np.sqrt(np.mean(np.sum(dif * dif, axis=1)))

    def report_statistics(self):
        """
        Returns a dictionary with statistics on the number of rotameric and non rotameric residues
        Ex: {'residue': 'PHE', 'rotamers': [4, 4, 5, 7], 'non_rotamers': 10}
        :return: dictionary
        """
        statistics = {'residue': self.residue_type.upper(), 'rotamers': [], 'non_rotamers': None}
        for cl in self.rotamer_classes[:-1]:
            statistics['rotamers'].append(len(cl))
        statistics['non_rotamers'] = len(self.rotamer_classes[-1])
        return statistics


def main():
    parser = argparse.ArgumentParser(description='Classify rotamers based on Penultimate rotamer library')
    parser.add_argument("-i", "--inp_directory", dest="inp_dir", required=True,
                        help="Directory with residues as separate files")
    parser.add_argument("-o", "--out_directory", dest="out_dir", required=True,
                        help="Output directory. If not given than the output will be saved to the input directory")
    parser.add_argument("-p", "--out_prefix", dest="prefix", default='rot_', help="Rotamer folders prefix")
    args = parser.parse_args()
    classify = PenultimateClassifier(args.inp_dir, args.out_dir, args.prefix)
    classify.run_classification()
    # print(classify.report_statistics())
    classify.log.info('%s\n', classify.report_statistics())


if __name__ == '__main__':
    main()



"""
mapMotifValidation.py

This module runs map versus amino acid residues map motif library validation.

Copyright [2013] EMBL - European Bioinformatics Institute
Licensed under the Apache License, Version 2.0 (the
"License"); you may not use this file except in
compliance with the License. You may obtain a copy of
the License at
http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied. See the License for the
specific language governing permissions and limitations
under the License.
"""

__author__ = 'Andrei Istrate'
__email__ = 'andrei@ebi.ac.uk'
__date__ = '2018-05-29'

import os
import argparse


def main():
    parser = argparse.ArgumentParser(description='Map Validation')
    parser.add_argument("-i", "--input", dest="inp", required=True, help="Strudel map motif validation output directory")
    parser.add_argument("-o", "--out", dest="out", required=True, help="Output directory")
    parser.add_argument("-log", "--log", dest="log", default=None, required=False, help="Log file")
    parser.add_argument("-v", "--voxel", dest="voxel", required=False, default=0.25, type=float, help="Segments voxel size")
    parser.add_argument("-r", "--recompute", dest="recompute_scores", action='store_true',
                        help="Recalculate correlations")
    parser.add_argument("-rs", "--recompute_segments", dest="recompute_segments", action='store_true',
                        help="Repeat segmenting of the input model and map if found in the output")
    parser.add_argument("-vc", "--verbose_chimerax", dest="v_c", action='store_true',
                        help="Run ChimeraX in verbose mode")
    parser.add_argument("-wl", "--warning_level", dest="warning_level", default='info',
                        help="Log file warning level [info[debug]")

    args = parser.parse_args()
    if args.log:
        log_file = args.log
    else:
        log_file = os.path.join(args.out, 'map_motif_validation.log')

    if not os.path.exists(args.out):
        os.makedirs(args.out)

    logging.basicConfig(filename=log_file, level=logging.INFO, format='%(levelname)s:  %(message)s')
    date_time = datetime.now().strftime("%H:%M:%S %Y-%m-%d")
    logging.info('\n{:_^100}'.format('MapMotifValidation') + '\n\nStarted: {}\n'.format(date_time))

    args.np = int(args.np)
    start = time.time()

    compute = ComputeScores()
    compute.set_paths(work_dir=args.out, lib=args.lib, in_map=args.in_map, in_model=args.in_model)
    chopped_pairs = compute.segment_structure(n_cores=args.np, voxel=args.voxel, replace=args.recompute_segments)
    json_file_path = compute.compute_correlations(chopped_pairs, args.np, args.recompute_scores, verbose=args.v_c)

    prefix = os.path.splitext(json_file_path)[0]
    if os.path.exists(prefix + '.csv'):
        csv_to_top_csv(prefix+'.csv', prefix+'_top.csv')
        # csv_to_top_scores_only_csv(prefix + '.csv', prefix + '_top_for_web.csv')

    logging.info('\nElapsed: {}\n{:_^100}'.format(func.report_elapsed(start), ''))


if __name__ == '__main__':
    main()


"""
chopModelMap.py

Class for Chopping an atomic residue (and EM map) into residues and residue side chains which are saved
as separate files.
The chopped side chains are used to chop the corresponding map.
This code uses the chopMap module for map chopping.

Copyright [2013] EMBL - European Bioinformatics Institute
Licensed under the Apache License, Version 2.0 (the
"License"); you may not use this file except in
compliance with the License. You may obtain a copy of
the License at
http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied. See the License for the
specific language governing permissions and limitations
under the License.
"""

__author__ = 'Andrei Istrate'
__email__ = 'andrei@ebi.ac.uk'
__date__ = '2018-05-29'


import copy
import os
import numpy as np
from datetime import datetime
import time
import json
from multiprocessing import Process, Array
try:
    from mpi4py import MPI
    mpi_avail = True
except ImportError:
    mpi_avail = False

from random import shuffle

from threed_strudel.chop.chopMap import ChopMap, MapParser
import threed_strudel.utils.functions as func
from threed_strudel.utils import bioUtils
import threed_strudel.nomenclature as nomenclature


# noinspection PyTypeChecker
class ChopModelMap:
    """
    Class for chopping residues and residue side chains out of proteins. The chopped
    side chains are used as guide for density map chopping
    """

    def __init__(self, rank=0, parallelism="shared"):
        self.rank = rank
        self.parallelism = parallelism
        if self.parallelism == "mpi" and mpi_avail:
            self.comm = MPI.COMM_WORLD
        else:
            self.comm = None
            self.parallelism = "shared"
        self.start_time = None
        self.min_rscc = None
        self.no_charged_check = None
        self.chop_log = None
        self.allowed_b = 0.0
        self.entry_code = ''
        self.work_dir = ''
        self.tmp_dir = None
        self.tmp_log_dir = None
        self.master_log_path = None
        self.input_model = None
        self.input_model_path = ''
        self.map_file_path = ''
        self.map_object = None
        self.rscc_list = []
        self.check_rscc = None
        self.side_chain_end = ''
        self.residue_end = ''
        self.shift_side_end = ''
        self.shift_residue_end = ''
        self.cube_res_end = ''
        self.cube_end = ''
        self.soft_map_end = ''
        self.hard_map_end = ''
        self.cube_radius = None
        self.final_voxel = None
        self.chop_radius = None
        self.chop_soft_radius = None
        # self.script_path = os.path.dirname(os.path.abspath(__file__))
        self.charged_res = nomenclature.CHARGED_RESIDUES
        # self.inclusion_threshold = None
        self.check_b_values = None
        # self.known_inclusion_levels = os.path.join(self.script_path, 'known_inclusion_levels.txt')
        self.inclusion_charged_tolerance = 3
        self.heavy_atoms = nomenclature.HEAVY_ATOMS

    def set_env(self, work_dir,
                entry_code,
                model_path,
                map_path,
                log_path,
                rscc_list=None,
                allowed_b=None,
                filter_identical_chains=True,
                warning_level='info'):
        """
        Set the environment for residue (and density map) chopping
        :param warning_level:
        :param inclusion_fraction: fraction of atoms inside the threshold
        :param allowed_b:
        :param rscc_list:
        :param log_path:
        :param work_dir: The directory for output files
        :param entry_code: EMD or pdb code of the entry
        :param model_path: Path to the residue pdb or cif file
        :param map_path: Path to the residue map file
        :param warning_level: warning level
        """
        self.start_time = time.time()
        s_env_t = time.time()
        self.work_dir = work_dir
        self.create_dir(work_dir)
        self.entry_code = entry_code
        self.input_model_path = model_path
        self.map_file_path = map_path
        self.master_log_path = log_path
        self.rscc_list = rscc_list

        self.tmp_log_dir = os.path.join(work_dir, 'tmp_log')
        self.create_dir(self.tmp_log_dir)
        tmp_log_path = os.path.join(self.tmp_log_dir, '{}{}{}'.format('rank_', self.rank, '.log'))

        if self.rank == 0:
            self.chop_log = func.setup_logger('chop_log', log_path, warning_level=warning_level)
            date_time = datetime.now().strftime("%H:%M:%S %Y-%m-%d")
            text = '\n{:_^100}'.format('ChopModelMap') + '\n\nStarted: {}'.format(date_time)
            self.chop_log.info('%s\n', text)
            self.chop_log.info('Reading the input files and preparing the environment..\n')
        else:
            if os.path.exists(tmp_log_path):
                os.remove(tmp_log_path)
            self.chop_log = func.setup_logger('chop_log', tmp_log_path, warning_level=warning_level)

        if self.parallelism == "mpi":
            if self.rank == 0:
                start = time.time()
                self.map_object = MapParser(map_path)
                self.chop_log.info('Map loaded in: %s', func.report_elapsed(start))
                self.input_model = bioUtils.load_structure(model_path)
                if filter_identical_chains:
                    self.input_model = self.delete_identical_chains(self.input_model)
            self.input_model = self.comm.bcast(self.input_model, root=0)
        else:
            self.map_object = MapParser(map_path)
            self.input_model = bioUtils.load_structure(model_path)
            if filter_identical_chains:
                self.input_model = self.delete_identical_chains(self.input_model)

        if allowed_b is None:
            self.allowed_b = self.calc_b_value_cut_off()
            if self.rank == 0:
                self.chop_log.info('\nNo B-factor cut-off was specified\n'
                                   'The cut-off value vas set to median + 2 * 1Qu: %s',
                                   self.allowed_b)

        self.set_out_file_suffixes()
        if self.rank == 0:
            self.chop_log.info("Spent on set_paths %s", func.report_elapsed(s_env_t))

    def set_map_chop_parameters(self, cube_radius=5.0, final_voxel=0.5, chop_radius=3.0, chop_soft_radius=2.0,
                                check_rscc=True, min_rscc=0.7, check_b_values=True, no_charged_check=True):
        """
        Set custom parameters for map chopping
        :param check_b_values:
        :param check_rscc:
        :param check_inclusion:
        :param cube_radius: The distance between the cube edge the residue
        :param final_voxel: Final voxel size of the chopped maps
        :param chop_radius: Radius around the molecule for map chopping
        :param chop_soft_radius: Soft hard_radius around the molecule for map chopping
        :param resolution:
        """
        self.cube_radius = cube_radius
        self.final_voxel = final_voxel
        self.chop_radius = chop_radius
        self.chop_soft_radius = chop_soft_radius
        self.check_rscc = check_rscc
        self.min_rscc = min_rscc
        self.check_b_values = check_b_values
        self.no_charged_check = no_charged_check

        # Set RSCC data
        if check_rscc and self.rscc_list is None:
            if self.rank == 0:
                raise Exception('Please provide the map resolution or a list of per residue RSCC')

    def set_out_file_suffixes(self):
        """
        Sets suffixes for output files
        """
        self.side_chain_end = '-{}_side.cif'.format(self.entry_code)
        self.shift_side_end = '-{}_side_shift.cif'.format(self.entry_code)
        self.residue_end = '-{}_residue.cif'.format(self.entry_code)
        self.shift_residue_end = '-{}_residue_shift.cif'.format(self.entry_code)
        self.cube_end = '-{}_cube.mrc'.format(self.entry_code)
        self.cube_res_end = '-{}_cube_resampled.mrc'.format(self.entry_code)
        self.soft_map_end = '-{}_soft.mrc'.format(self.entry_code)
        self.hard_map_end = '-{}_hard.mrc'.format(self.entry_code)

    def delete_identical_chains(self, structure, delta=0.15):
        """
        Deletes duplicating chains in a structure residue based on pairwise RMS (delta)
        :param delta: maximum pairwise RMS to consider chains identical
        :param structure: Biopython structure object
        :return: Biopython structure object
        """
        # if structure.level == 'S':
        #     structure = structure[0]
        chain_ids = [chain.get_id() for chain in structure.get_chains()]
        if self.rank == 0:
            start = time.time()
            self.chop_log.info('\nSearching identical chains..')
            self.chop_log.info('All chains:')
            self.chop_log.info(' '.join(chain_ids) + '\n')

        classes, max_distances = bioUtils.classify_chains(structure, delta=delta)
        duplicates = [cl for cl in classes if len(cl) > 1]
        # if self.rank == 0:
        #     self.chop_log.info(' '.join(classes))
        if len(duplicates) > 0:
            if self.rank == 0:
                self.chop_log.info('Identical chains:')
                for cl in duplicates:
                    self.chop_log.info(' '.join(cl))

            unique_best_chains = bioUtils.select_best_chains(structure, self.map_object, classes)
            if self.rank == 0:
                self.chop_log.info('Selected as best fitting in the map unique chains:')
                self.chop_log.info(' '.join(unique_best_chains))
            for chain_id in chain_ids:
                if chain_id not in unique_best_chains:
                    del structure[0][chain_id]
        elif self.rank == 0:
            self.chop_log.info("No identical chains were found")
            self.chop_log.info('Elapsed: %s', func.report_elapsed(start))
        return structure

    @staticmethod
    def get_residue_list(structure, res):
        """
        Create a list of BIO.PDB residue objects containing all the residues
        of the given type in the input residue
        :param structure: BIO.PDB structure object
        :param res: Residue code
        :return: A list of BIO.PDB residue objects
        """
        del_atom_list = []

        for model in structure:
            for chain in model:
                for residue in chain:
                    for atom in residue:
                        if atom.get_name().upper().startswith('H'):
                            del_atom_list.append((chain.get_id(), residue.get_id(), atom.get_id()))
        for chain, resi, atom in del_atom_list:
            del structure[0][chain][resi][atom]
        residues = []
        for residue in structure.get_residues():
            if residue.get_resname() == res.upper():
                residues.append(residue)

        return residues

    def calc_b_value_cut_off(self):
        """
        Computes B-factor statistics
        :return: median + 2 * 1st Quartile
        """
        b = []
        for atom in [a for a in self.input_model.get_atoms()]:
            if atom.parent.resname.rstrip() != 'HOH':
                b.append(atom.bfactor)
        np_b = np.array(b)
        median = np.median(np_b)
        qu_1 = np.percentile(np_b, 25, interpolation='lower')
        b_cut = median + 2 * qu_1

        return int(b_cut)

    def check_b_factors(self, residue):
        """
        Analyse the B-factors of a BIO.PDB residue object in order to asses the local quality
        of the density map. Returns true if the B-factors are consistent with high quality map.
        :param residue: BIO.PDB residue object
        :return: True or False
        """
        valid = True
        for atom in residue:
            if atom.bfactor > self.allowed_b:
                valid = False
                break
        return valid

    def run_model_quality_checks(self, residue, chop_map):
        """
        Runs the residue residue quality checks
        :param residue: residue object
        :param chop_map: map dir
        :return: False or True
        """
        good_b = True
        good_rscc = True
        complete = True
        chain_id = residue.parent.id
        res_nr = residue.id[1]
        res_name = residue.get_resname()

        nr_atoms = 0
        single_conf = True
        for atom in residue:
            if atom.is_disordered():
                single_conf = False
            #     break
            # try:
            #     occupancy = atom.last_occupancy
            # except AttributeError:
            #     occupancy = 1
            # if occupancy < 1:
            #     single_conf = False
                self.chop_log.info('The %s %s %s residue has multiple conformations',
                                   residue.get_resname().upper(), residue.id[1], residue.parent.id)
                break
            nr_atoms += 1

        # Check heavy atom completeness
        if nr_atoms < self.heavy_atoms[res_name]:
            self.chop_log.info('Missing heavy atoms in %s %s %s (%s out of %s)', res_name,
                               res_nr, chain_id, nr_atoms, self.heavy_atoms[res_name])
            complete = False
        elif nr_atoms > self.heavy_atoms[res_name]:
            self.chop_log.info('Terminal residue will be skipped %s %s %s (%s out of %s)', res_name,
                               res_nr, chain_id, nr_atoms, self.heavy_atoms[res_name])
            complete = False

        # Check local map quality based on B-factors and atom inclusion
        if chop_map:
            if res_name.upper() not in self.charged_res or not self.no_charged_check:
                # Check B-factors
                if self.check_b_values:
                    good_b = self.check_b_factors(residue)
                    if not good_b:
                        self.chop_log.info('Residue %s %s %s has high B-factors', res_name, res_nr, chain_id)
                # Check map residue real space correlation coefficients
                if self.check_rscc:
                    rscc = self.find_residue_rscc(self.rscc_list, chain_id, res_name, res_nr)
                    if rscc < self.min_rscc:
                        self.chop_log.info('Residue %s %s %s has RSCC lover than %s',
                                           res_name, res_nr, chain_id, self.min_rscc)
                        good_rscc = False

        if good_rscc and good_b and complete and single_conf:
            return True
        else:
            return False

    @staticmethod
    def find_residue_rscc(rscc_list, chain, res_type, res_nr):
        """
        Find the residue RSCC in the list
        :param rscc_list:
        :param chain:
        :param res_type:
        :param res_nr:
        :return:
        """
        result = None
        for residue in rscc_list:
            if residue[0] == chain and residue[1] == res_type and int(residue[2]) == res_nr:
                result = residue[3]
                break
        return result

    def build_all_residue_object_list(self, residue_list):
        """
         Create a list of BIO.PDB residue objects containing all the residues
         to be chopped in the input residue
         :param residue_list: list of residues names
         :return: A list of BIO.PDB residue objects
         """
        all_res_obj_list = []
        for res in residue_list:
            # Create a list of all current type residue objects
            residues = self.get_residue_list(self.input_model, res)
            all_res_obj_list = all_res_obj_list + residues
        all_res_obj_list.sort(key=lambda z: z.get_resname())

        return all_res_obj_list

    def create_dir_structure(self, residue_list):
        """
        Creates the chopping directory structure
        :param residue_list: list of residue names
        """
        for res in residue_list:
            res_dir = self.work_dir + '/' + res.lower()
            self.create_dir(res_dir)
            side_chain_dir = res_dir + '/' + 'side_chain'
            self.create_dir(side_chain_dir)
            model_map_dir = res_dir + '/' + 'models_and_maps'
            self.create_dir(model_map_dir)
            lq_model_map_dir = res_dir + '/' + 'lq_models_and_maps'
            self.create_dir(lq_model_map_dir)

    def chop_worker_task(self, residue_obj_list, in_map, stat_storage,
                         chop_map=True, chopping_mode='soft', save_tmp_files=False):
        """
        This function is a wrapper for the code which will be executed in parallel
        :param stat_storage: list of containers for chopping statistics
        :param in_map: input map as MapObject or map file in_dir
        :param self: The outer scape self object
        :param residue_obj_list: list of residues objects
        :param chop_map: chop both residue and map
        :param chopping_mode: 'soft' or 'hard'
        """
        chop = ChopMap()
        resampling_time = 0
        res_order_dict = stat_storage[0]
        hq = stat_storage[1]
        lq = stat_storage[2]
        tmp_dir = '/tmp/chop'
        self.tmp_dir = tmp_dir
        try:
            os.makedirs(tmp_dir)
        except FileExistsError:
            pass
        map_obj_list, shifts_list = chop.chop_cube_list(residue_obj_list,
                                                        in_map,
                                                        self.cube_radius)

        for i in range(len(residue_obj_list)):
            c_map_obj = map_obj_list[i]
            residue = residue_obj_list[i]
            matrix = shifts_list[i]

            res_type = residue.get_resname().lower()
            chain_id = residue.parent.id
            res_nr = residue.id[1]
            res_dir = self.work_dir + '/' + res_type

            # Run local residue quality checks
            good_model = self.run_model_quality_checks(residue, chop_map)
            for atom in residue:
                atom.disordered_flag = 0
            # Increment ctype objects
            if good_model:
                hq[res_order_dict[res_type]] = hq[res_order_dict[res_type]] + 1
                model_map_dir = res_dir + '/' + 'models_and_maps'
            else:
                lq[res_order_dict[res_type]] = lq[res_order_dict[res_type]] + 1
                model_map_dir = res_dir + '/' + 'lq_models_and_maps'
            side_pdb_dir = res_dir + '/' + 'side_chain'
            # Save the residue as pdb file
            name_prefix = '{}-{}-{}'.format(res_type, res_nr, chain_id)
            # Delete the main chain atoms
            side_chain = residue
            tmp_res_path = os.path.join(tmp_dir, name_prefix + '.cif')
            struct = bioUtils.residues2structure(residue)
            bioUtils.save_model(struct, tmp_res_path)
            residue = bioUtils.load_structure(tmp_res_path)
            os.remove(tmp_res_path)
            side_chain = self.del_main_chain(side_chain)

            # Save the side chain
            cube_map_path = os.path.join(tmp_dir, name_prefix + self.cube_end)
            if save_tmp_files:
                c_map_obj.write_map(cube_map_path)
            # Resampled map path
            cube_new_grid_path = os.path.join(tmp_dir, name_prefix + self.cube_res_end)
            start = time.time()

            # chop.grid_resample(cube_map_path, cube_new_grid_path, self.final_voxel)

            c_map_obj.grid_resample_emda(self.final_voxel)
            if save_tmp_files:
                c_map_obj.write_map(cube_new_grid_path)

            end = time.time()
            resampling_time += (end - start)
            bioUtils.shift_coord(matrix, side_chain)
            if chopping_mode.lower() == 'hard':
                fin_map = os.path.join(model_map_dir, name_prefix + self.hard_map_end)
                # chop.chop_hard_radius(side_chain, cube_new_grid_path, fin_map, self.chop_radius)
                chop.chop_hard_radius(side_chain, c_map_obj, fin_map, self.chop_radius)
            elif chopping_mode.lower() == 'soft':
                fin_map = os.path.join(model_map_dir, name_prefix + self.soft_map_end)
                # chop.chop_soft_radius(side_chain, cube_new_grid_path, fin_map, self.chop_radius,
                #                       self.chop_soft_radius)
                chop.chop_soft_radius(side_chain, c_map_obj, fin_map, self.chop_radius,
                                      self.chop_soft_radius)
            else:
                raise Exception('The chopping_mode parameter can be hard or soft')

            # try:
            #     os.remove(cube_map_path)
            #     os.remove(cube_new_grid_path)
            # except FileNotFoundError:
            #     pass

            fin_res = os.path.join(model_map_dir, name_prefix + self.shift_residue_end)
            side_path = os.path.join(side_pdb_dir, name_prefix + self.side_chain_end)
            side_struct = bioUtils.residues2structure(side_chain)
            bioUtils.save_model(side_struct, side_path)
            bioUtils.shift_coord(matrix, residue)
            # res_struct = bioUtils.residues2structure(residue)
            bioUtils.save_model(residue, fin_res)
        self.chop_log.debug("Time spent on resampling: %s s.", resampling_time)
        if self.rank != 0:
            self.comm.send([res_order_dict, hq, lq], dest=0, tag=1)
        return [res_order_dict, hq, lq]

    def chop_model_map_parallel(self, residue_list, chop_map=True, chopping_mode='soft',
                                nr_cores=2):
        """
        Similar to chop_model_map but runs in parallel
        Chop the specified residues and their side chains out of the input residue. Each residue and
        the corresponding side chain is saved in separate pdb files. The residues are classified as
        having high quality (highq_residue) or low quality (lowq_residue) local density map,
        superimposed and saved in multimodel pdb files.
        :param nr_cores: Number of cores
        :param residue_list: The list of residue type to be chopped (ex. [ASP, TYR, ARG]).
        :param chop_map: Boolean. To perform map chopping or not.
        :param chopping_mode: 'soft' or 'hard'
        """
        start = time.time()
        chop = ChopMap()
        residue_list.sort()
        all_res_obj_list = []

        if self.rank == 0:
            # Cut the map around initial residue to save memory
            map_obj_list, shifts_l = chop.chop_cube_list([self.input_model], self.map_object, 4)
            if not map_obj_list:
                return None

            bioUtils.shift_coord(shifts_l[0], self.input_model)

            # Update the main map object
            self.map_object = map_obj_list[0]

            # Save the residue and map which will be actually used for chopping
            bioUtils.save_model(self.input_model, os.path.join(self.work_dir, self.entry_code + '_shifted.cif'))
            cut_map_path = os.path.join(self.work_dir, self.entry_code + '_model_part.mrc')
            self.map_object.write_map(cut_map_path)
            self.map_file_path = cut_map_path

            # Get the lists of residues for chopping
            all_res_obj_list = self.build_all_residue_object_list(residue_list)
            split_all_res_obj_list = func.split_sorted_list(all_res_obj_list, nr_cores)
            for sublist in split_all_res_obj_list:
                shuffle(sublist)
            # Create the chopping directory structure
            self.create_dir_structure(residue_list)

        else:
            split_all_res_obj_list = None

        # Dictionary to track the residue position in the list ("ASP": index)
        res_order_dict = {}
        for index, r in enumerate(residue_list):
            res_order_dict[r.lower()] = index
        if self.parallelism == "mpi":
            if self.rank == 0:
                hq = [0 for _ in range(len(residue_list))]
                lq = [0 for _ in range(len(residue_list))]
                stat_storage = []
                for i in range(nr_cores):
                    stat_storage.append(copy.deepcopy([res_order_dict, hq, lq]))
            else:
                stat_storage = None
        else:
            # Create ctype objects which are available to all Processes
            hq = Array('i', [0 for _ in range(len(residue_list))])
            lq = Array('i', [0 for _ in range(len(residue_list))])
            stat_storage = [res_order_dict, hq, lq]

        if self.rank == 0:
            self.chop_log.info('Spent on prep in chop_model_map: %s', func.report_elapsed(start))
            self.chop_log.info('All ready for chopping\nElapsed: %s', func.report_elapsed(self.start_time))
            if self.parallelism == 'mpi':
                mem = 'distributed'
            else:
                mem = 'shared'
            self.chop_log.info('\nStarting chopping on %s cores with %s memory parallelism..', nr_cores, mem)

            txt = ''
            for i, r in enumerate(residue_list):
                if i % 10 == 0 and i > 0:
                    txt = txt + '\n          '
                else:
                    txt = txt + r + ' '

            self.chop_log.info('Chopping: %s', txt)
            self.chop_log.info('%s residues to chop', len(all_res_obj_list))

        # Setup and run parallel processes
        start = time.time()

        if self.parallelism == 'mpi':
            if self.rank == 0:
                self.chop_log.info('\nBroadcasting data..')
            # distributed memory
            split_all_res_obj_list = self.comm.bcast(split_all_res_obj_list, root=0)
            self.map_object = self.comm.bcast(self.map_object, root=0)
            self.input_model = self.comm.bcast(self.input_model, root=0)
            self.map_file_path = self.comm.bcast(self.map_file_path, root=0)
            # self.inclusion_threshold = self.comm.bcast(self.inclusion_threshold, root=0)
            stat_storage = self.comm.scatter(stat_storage, root=0)
            self.chop_log.info("\nRank: %s spent %s on broadcasting", self.rank, func.report_elapsed(start))
            start = time.time()
            # Run
            stat_storage = self.chop_worker_task(split_all_res_obj_list[self.rank],
                                                 self.map_object,
                                                 stat_storage,
                                                 chop_map=chop_map,
                                                 chopping_mode=chopping_mode)
            self.chop_log.info('Rank %s finished chopping of %s residues in %s',
                               self.rank, len(split_all_res_obj_list[self.rank]), func.report_elapsed(start))
            self.chop_log.info('%s', '{:*^80}'.format(''))
            stat_storage = self.comm.gather(stat_storage, root=0)
        else:
            # shared memory
            thread_list = []
            for residue_obj_list in split_all_res_obj_list:
                t = Process(target=self.chop_worker_task,
                            args=(residue_obj_list, self.map_object, stat_storage, chop_map, chopping_mode))
                thread_list.append(t)
            for thread in thread_list:
                thread.start()
            for thread in thread_list:
                thread.join()
            self.chop_log.info('Finished chopping\nElapsed: %s', func.report_elapsed(start))
            self.chop_log.info('%s', '\n{:*^80}\n'.format(''))

        # Build a dictionary to store chopping statistics ( {res_name: [hq_nr, lq_nr]} )
        if self.rank == 0:
            # Collect the statistics from all cores
            if self.parallelism == "mpi":
                tmp = stat_storage[0]
                for storage in stat_storage[1:]:
                    for i in range(len(tmp[1])):
                        tmp[1][i] += storage[1][i]
                        tmp[2][i] += storage[2][i]
                stat_storage = tmp
            hq = stat_storage[1]
            lq = stat_storage[2]

            # Add workers logs to the master log
            with open(self.master_log_path, 'a') as master_log:
                tmp_files = os.listdir(self.tmp_log_dir)
                for file in tmp_files:
                    tmp_file_path = os.path.join(self.tmp_log_dir, file)
                    with open(tmp_file_path, 'r') as tmp_log:
                        tmp_content = tmp_log.read()
                    master_log.write(tmp_content)

            statistics_dict = {}
            for i in range(len(residue_list)):
                statistics_dict[residue_list[i]] = [hq[i], lq[i]]

            text = self.format_statistics(statistics_dict)
            self.chop_log.info(text)

            date_time = datetime.now().strftime("%H:%M:%S %Y-%m-%d")
            self.chop_log.info("\nChopModelMap process successfully finished  %s", date_time)
            self.chop_log.info("Elapsed: %s", func.report_elapsed(self.start_time))
            self.chop_log.info('%s', '{:_^100}'.format(''))
            return statistics_dict
        else:
            # try:
            #     rmtree(self.tmp_dir)
            # except FileNotFoundError:
            #     pass
            return None

    @staticmethod
    def format_statistics(statistics_dict):
        """
        Formats the chopping statistics for the log output
        :param statistics_dict: Choping statistics dictionary ex: {res_name: [hq_nr, lq_nr]}
        :return: Formatted text
        """
        line1 = '{:<10}'.format('Residue')
        line2 = '{:<10}'.format(' High Q')
        line3 = '{:<10}'.format(' Low  Q')
        total_hq = 0
        total_lq = 0
        for key in sorted(statistics_dict.keys()):
            line1 = line1 + '{:>6}'.format(key)
            line2 = line2 + '{:>6}'.format(statistics_dict[key][0])
            line3 = line3 + '{:>6}'.format(statistics_dict[key][1])
            total_hq = total_hq + statistics_dict[key][0]
            total_lq = total_lq + statistics_dict[key][1]

        text = '\nResidue quality statistics:\n{}\n{}\n{}'.format(line1, line2, line3)
        text2 = '\n\n{:<18}{}\n{:<18}{}\n{:<18}{}'.format('Total:', total_hq + total_lq, 'High quality:', total_hq,
                                                          'Low quality:', total_lq)
        text = text + text2
        return text

    @staticmethod
    def del_main_chain(residue):
        """
        Delete main chain atoms
        :param residue: BIO.PDB object
        :return: BIO.PDB object
        """
        for atom in residue:
            if atom.get_name() == "O":
                del residue[atom.get_name()]
        for atom in residue:
            if atom.get_name() == "C":
                del residue[atom.get_name()]
        # for atom in residue:
        #    if atom.get_name() == "CA":
        #        del residue[atom.get_name()]
        for atom in residue:
            if atom.get_name() == "N":
                del residue[atom.get_name()]
        return residue

    @staticmethod
    def create_dir(path):
        """
        Create new directory
        :param path: new directory in_dir
        """
        try:
            os.makedirs(path)
        except FileExistsError:
            pass

import argparse
import sys


def main():
    parser = argparse.ArgumentParser(description='Chops a map and an atomic model into residues and corresponding maps')
    parser.add_argument("-p", "--pdb_cif", dest="in_model", required=True, help="Input model in pdb or cif format")
    parser.add_argument("-m", "--map", dest="in_map", required=True, help="Input map")

    parser.add_argument("-o", "--chop_dir", dest="chop_dir", required=True, help="Where to output the generated files")
    parser.add_argument("-c", "--chop_par", dest="chop_par", nargs='*', required=False,
                        help="Chopping parameters. Example: cube_radius=4 final_voxel=0.5 chop_radius=2.0 "
                             "chop_soft_radius=1.0 inclusion_fraction=90 chop_map=True chopping_mode=soft "
                             "check_rscc=True")
    parser.add_argument("-r", "--rscc_file", dest="rscc_file", required=False,
                        help='Per residue RSCC values in json format EX: [ ["A", "MET", "1", 0.827],..]')
    parser.add_argument("-re", "--resolution", dest="res", required=False,
                        help="Map resolution required if RSCC were not provided")
    parser.add_argument("-rl", "--residue_list", dest="res_list", required=False, nargs='*',
                        help="Residue names to be chopped Ex: ASP ARG ..")
    parser.add_argument("-np", "--n_processors", dest="np", required=False,
                        help="Number of processors in chopping mode")
    parser.add_argument("-mpi", "--mpi", dest="mpi", action='store_true',
                        help="Use distributed memory parallelism")
    parser.add_argument("-filter_chain", "--filter_chain", dest="filter", action='store_true', default=False,
                        help="Do not chop identical chains")

    default_parameters = {"cube_radius": 4, "final_voxel": 0.25, "chop_radius": 2.0, "chop_soft_radius": 1.0,
                          "chop_map": True, "chopping_mode": "soft", "check_rscc": True}
    args = parser.parse_args()
    if args.mpi:
        sys.excepthook = func.global_except_hook
        comm = MPI.COMM_WORLD
        size = comm.Get_size()
        rank = comm.Get_rank()
    else:
        rank = 0
        size = 1

    #config = Configure()
    chop = ChopModelMap(rank, parallelism='mpi')
    chop_dir = args.chop_dir

    if args.rscc_file:
        with open(args.rscc_file, 'r') as j:
            rscc_list = json.load(j)
    else:
        rscc_list = None

    # Set chopping parameters
    parameters = {}
    if args.chop_par:
        for parameter in args.chop_par:
            pair = parameter.split('=')
            if pair[1].lower == 'true':
                pair[1] = True
            if pair[1].lower() == 'false':
                pair[1] = False
            if pair[0] not in ['chop_map', 'chopping_mod', 'check_rscc', 'chop_map']:
                pair[1] = float(pair[1])
            parameters[pair[0]] = pair[1]

        for key, value in default_parameters.items():
            if key not in parameters.keys():
                parameters[key] = value
    else:
        parameters = default_parameters

    if args.res_list:
        res_list = args.res_list
    else:
        res_list = nomenclature.AA_RESIDUES_LIST

    if args.np:
        np = args.np
    else:
        np = 2
    if args.res:
        args.res = float(args.res)

    model_name = os.path.basename(args.in_model).split('.')[0]
    map_name = os.path.basename(args.in_map).split('.')[0]
    name = str(model_name) + '-' + str(map_name)
    log_path = os.path.join(chop_dir, 'chop_' + name + '.log')

    chop.set_env(args.chop_dir, name, args.in_model, args.in_map, log_path,
                 rscc_list=rscc_list,
                 allowed_b=None,
                 warning_level='debug',
                 filter_identical_chains=args.filter)
    chop.set_map_chop_parameters(cube_radius=parameters["cube_radius"],
                                 final_voxel=parameters["final_voxel"],
                                 chop_radius=parameters["chop_radius"],
                                 chop_soft_radius=parameters["chop_soft_radius"],
                                 check_rscc=parameters["check_rscc"],
                                 )

    statistics = chop.chop_model_map_parallel(res_list, chop_map=parameters["chop_map"],
                                              chopping_mode='soft',
                                              nr_cores=size)

if __name__ == '__main__':
    main()


"""
The Penultimate Rotamer Library
Simon C. Lovell, J. Michael Word, Jane S. Richardson, and David C. Richardson
PROTEINS: Structure, Function, and Genetics 40:389–408 (2000)
"""
import json
import os


basedir = os.path.dirname(os.path.abspath(__file__))
with open(os.path.join(basedir, 'penultimate_lib.json')) as j:
    PENULTIMATE_2000 = json.load(j)



"""
modelMapUtils.py

A collection of functions for atomic models and density maps manipulations

Copyright [2013] EMBL - European Bioinformatics Institute
Licensed under the Apache License, Version 2.0 (the
"License"); you may not use this file except in
compliance with the License. You may obtain a copy of
the License at
http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied. See the License for the
specific language governing permissions and limitations
under the License.
"""

__author__ = 'Andrei Istrate'
__email__ = 'andrei@ebi.ac.uk'
__date__ = '2019-10-09'

import numpy as np
from scipy.interpolate import RegularGridInterpolator


def interpolator(map_object):
    """
    Setup scipy regular grid interpolation function
    :return: interpolation function
    """
    nx, ny, nz = map_object.data.shape
    x = range(nx)
    y = range(ny)
    z = range(nz)
    _interpolator = RegularGridInterpolator((x, y, z), map_object.data)
    return _interpolator


def atom_inclusion(structure, map_object, threshold, ignore_atoms=()):
    """
    Counts the number of atoms within and outside map at a given level
    :param threshold:
    :param map_object:
    :param structure:
    :param ignore_atoms:
    :return: [included(nr), not_included(nr), fraction]
    """
    if structure.get_level() == 'S':
        model = structure[0]
    else:
        model = structure

    ignore_atoms = [atom.lower() for atom in ignore_atoms]

    _interpolator = interpolator(map_object)

    included = 0
    not_included = 0
    for atom in model.get_atoms():
        if atom.get_name().lower() not in ignore_atoms:
            map_index = map_object.coord_to_index(atom.coord)
            try:
                val = _interpolator(map_index)
            except ValueError:
                val = threshold

            if val > threshold:
                included += 1
            else:
                not_included += 1
        else:
            included += 1

    fraction = included / (included + not_included)

    return included, not_included, fraction


def find_threshold(structure, map_object, inclusion_fraction, delta=2):
    """
    Calculates the threshold level for which the specified fraction of atoms is inside the map
    :param map_object:
    :param structure:
    :param inclusion_fraction: atom inclusion fraction (%)
    :param delta: atom inclusion fraction precision
    :return: threshold level
    """
    _interpolator = interpolator(map_object)
    upper = np.nanmax(map_object.data)
    lower = np.nanmin(map_object.data)
    values = []
    for atom in structure.get_atoms():
        map_index = map_object.coord_to_index(atom.coord)
        try:
            val = _interpolator(map_index)
        except ValueError:
            val = lower
        values.append(val)
    values = np.array(values)
    total_atoms = len(values)
    current_lvl = 0
    final_lvl = 0.
    delta_lvl = upper - lower
    while delta_lvl > (upper - lower) / 100:
        tmp = np.where(values > current_lvl)
        included = len(values[tmp])
        not_included = total_atoms - included

        current_incl = included / (included + not_included) * 100
        if current_lvl == 0 and current_incl < inclusion_fraction:
            final_lvl = 0.
            break
        elif current_incl < inclusion_fraction:
            upper = current_lvl
            delta_lvl = (upper - lower) / 2
            current_lvl = current_lvl - delta_lvl
        elif current_incl > inclusion_fraction + delta:
            lower = current_lvl
            delta_lvl = (upper - lower) / 2
            current_lvl = current_lvl + delta_lvl
        else:
            final_lvl = round(current_lvl, 8)
            break

    return final_lvl


def coord_to_index(coord, map_obj):
    """
    Converts atomic residue coordinates into map indices
    :param coord: coordinate in angstrom
    :param map_obj: map object
    :return: float indices
    """
    st = map_obj.n_start
    vs = map_obj.voxel_size
    orig = map_obj.origin

    x_index = coord[2] / vs[2] - orig[2] / vs[2] - st[2]
    y_index = coord[1] / vs[1] - orig[1] / vs[1] - st[1]
    z_index = coord[0] / vs[0] - orig[0] / vs[0] - st[0]
    return x_index, y_index, z_index


def coord_to_index_(coord, map_obj):
    """
    Converts atomic residue coordinates into map indices
    :param coord: coordinate in angstrom
    :param map_obj: map object
    :return: float indices
    """
    return coord[2] / map_obj.voxel_size[2] - map_obj.origin[2] / map_obj.voxel_size[2] - map_obj.n_start[2], \
           coord[1] / map_obj.voxel_size[1] - map_obj.origin[1] / map_obj.voxel_size[1] - map_obj.n_start[1], \
           coord[0] / map_obj.voxel_size[0] - map_obj.origin[0] / map_obj.voxel_size[0] - map_obj.n_start[0]


def coord_to_index_int(coord, map_obj):
    """
    Converts atomic residue coordinates into map indices
    :param coord: coordinate in angstrom
    :param map_obj: map object
    :return: float indices
    """
    st = map_obj.n_start
    vs = map_obj.voxel_size
    orig = map_obj.origin

    x_index = int(round(coord[2] / vs[2] - orig[2] / vs[2] - st[2]))
    y_index = int(round(coord[1] / vs[1] - orig[1] / vs[1] - st[1]))
    z_index = int(round(coord[0] / vs[0] - orig[0] / vs[0] - st[0]))
    return x_index, y_index, z_index



from subprocess import check_output
import time
import subprocess
import os


LSF_TEMPLATE = 'bsub -o {}.out -e {}-err.out -M {} -n {} '


def memory_supervised_lsf_run(job, prefix, workdir,
                              start_mem=1000, mem_delta=1000, mem_limit=80000, np=1):
    """
    Submits a job to the cluster LSF queue system. If the job is killed due
    to exceeded memory then the memory limit is increased.
    :param np: number of cores
    :param start_mem: default memory
    :param job: job command or bash file be run on cluster
    :param prefix: a string with the name of the program (used as prefix for stderr and stdout)
    :param workdir: directory where the job will be run
    :param mem_delta: memory increase steep
    :param mem_limit: maximum available memory
    """

    def run_monitored_job(command, work_dir):
        """
        Submits a job to the queue system and waits until it finishes
        :param command: Bsub command
        :param work_dir: directory where the job will be run
        """
        command = command.split()
        # Submit job to the queue ang get the job id
        out = str(check_output(command, cwd=work_dir))
        job_id = out.split('<')[1].split('>')[0]
        while True:
            # Check the job status every 20 seconds
            try:
                out1 = str(check_output(["bjobs", "-a", "-noheader", job_id]))
                status = out1.split()[2]
            except subprocess.CalledProcessError:
                continue
            if status == 'DONE' or status == 'EXIT':
                break
            elif 'not found' in out1:
                break
            else:
                time.sleep(20)

    def read_max_used_mem(bsub_out):
        """
        Reads the maximum and exceeded amount of memory during the last bsub run.
        :param bsub_out: bsub output file
        :return: maximum used memory, exceeded amount
        """
        memterm = False
        last = -1
        maxmem = 0
        try:
            with open(bsub_out) as f:
                lines = f.readlines()
                for nr, line in enumerate(lines):
                    if 'LSBATCH' in line:
                        last = nr
                for line in lines[last:]:
                    if 'TERM_MEMLIMIT' in line:
                        memterm = True
                if memterm:
                    for line in lines[last:]:
                        try:
                            if 'Max Memory' in line:
                                maxmem = int(float(line.split()[-2]))
                        except ValueError:
                            pass
        except IOError:
            pass
        return memterm, maxmem

    bsub_prefix = build_bsub_command(prefix, start_mem, np)
    jobstring = bsub_prefix + ' ' + job
    run_monitored_job(jobstring, workdir)
    current_mem = start_mem
    while True:
        # Check the job summary. If the memory limit was exceeded rerun with more memory
        mem_term, max_mem = read_max_used_mem(os.path.join(workdir, prefix + '.out'))
        if not mem_term:
            break
        elif max_mem > mem_limit:
            break
        elif max_mem == 0:
            current_mem += mem_delta
        elif max_mem < mem_limit:
            current_mem = max_mem + mem_delta

        bsub_prefix = build_bsub_command(prefix, current_mem, np)
        jobstring = bsub_prefix + ' ' + job
        run_monitored_job(jobstring, workdir)
    # Sometimes a file created by a node on cluster is not immediately visible to other cores
    # which results in os.in_dir.exists returning False on existing files. Waiting a bit helps
    time.sleep(10)
    return current_mem


def build_bsub_command(prefix, memory, np=1):
    """
    Creates the cluster job submission command
    Specific to LSF platform (needs to be modified for other cluster platforms
    :param np: number of cores
    :param prefix: prefix for the LSF output files
    :param memory: Memory limit
    :return: LSF command
    """
    command = LSF_TEMPLATE.format(prefix, prefix, memory, np)
    return command


"""
chopMap.py

A collection of methods which allow to chop a map around an atomic model.
The map can be chopped in three different ways:
- using a cub around the atomic residue with hard edges
- using certain radius around atomic residue with hard edges
- using certain radius around atomic residue with soft edges

Copyright [2013] EMBL - European Bioinformatics Institute
Licensed under the Apache License, Version 2.0 (the
"License"); you may not use this file except in
compliance with the License. You may obtain a copy of
the License at
http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied. See the License for the
specific language governing permissions and limitations
under the License.
"""

__author__ = 'Andrei Istrate'
__email__ = 'andrei@ebi.ac.uk'
__date__ = '2018-05-29'

import math
from random import randint
import mrcfile
import numpy as np

from threed_strudel.parse.mapParser import MapParser
from threed_strudel.utils import bioUtils


class ChopMap:
    """
    Class for chopping map around an atomic model
    """

    def __init__(self, in_model=None, in_map=None):
        self.model = in_model
        self.map = in_map
        self.sett_model_map()

    def sett_model_map(self):
        """
        Loads the map and model
        """
        if self.map is not None:
            self.map = MapParser(self.map)
        if self.model is not None:
            self.model = bioUtils.load_structure(self.model)

    @staticmethod
    def chop_cube_list(in_model, in_map, cube_padding, zero_origin=True):
        """
         Chop map using a cubic box around the model
         :param in_model: biopython model object or list of objects
         :param in_map: path to the input map or strudel map object
         :param cube_padding: distance fom any cube edge to the nearest atom
         :param zero_origin: boolean,
         :return: map object or list of map objects, translation matrix or list of matrices depending on input 
         """
        lst = True
        if type(in_map) is str:
            in_map = MapParser(in_map)
        if type(in_model) is not list and type(in_model) is not tuple:
            in_model = [in_model]
            lst = False

        shifts_list = []
        chopped_map_obj_list = []

        for model in in_model:

            atoms_coord = []
            for atom in model.get_atoms():
                atoms_coord.append(atom.coord)

            delta = round(cube_padding / in_map.voxel_size[0])
            # Get the indices in the map grid that correspond to the atom coordinates
            x_indices = []
            y_indices = []
            z_indices = []
            for atom in atoms_coord:
                x_index, y_index, z_index = in_map.coord_to_index(atom)
                if any([x_index, y_index, z_index]) < 0:
                    print("Atom outside map")
                else:
                    x_indices.append(x_index)
                    y_indices.append(y_index)
                    z_indices.append(z_index)
            # Find the voxel located in the middle of the atomic residue
            # and the maximum molecule size in grid points
            deltas = []
            minim = min(x_indices)
            dx = max(x_indices) - minim
            middle_x = int(round(dx / 2 + minim))
            deltas.append(dx)

            minim = min(y_indices)
            dy = max(y_indices) - minim
            middle_y = int(round(dy / 2 + minim))
            deltas.append(dy)

            minim = min(z_indices)
            dz = max(z_indices) - minim
            middle_z = int(round(dz / 2 + minim))
            deltas.append(dz)
            max_d = max(deltas)
            # Calculate the size of the cube
            radius = int(round(max_d / 2 + delta))
            new_dimension = radius * 2
            # Ensure that the new grid size has no prime numbers greater than 19

            # new_dimension = func.find_good_grid([new_dimension, new_dimension, new_dimension])[0]
            # print(new_dimension)
            # hard_radius = int(new_dimension / 2)
            # Create a numpy array to store the chopped voxels
            new_data = np.zeros((new_dimension, new_dimension, new_dimension), dtype='float32')

            # Assign voxel values
            for x in range(new_dimension):
                for y in range(new_dimension):
                    for z in range(new_dimension):
                        try:
                            new_data[x, y, z] = in_map.data[x + middle_x - radius, y + middle_y - radius,
                                                            z + middle_z - radius]
                        except IndexError:
                            pass
            # Calculate the new cell size
            voxel_size = in_map.voxel_size
            new_cell = (round(new_dimension * voxel_size[0], 3),
                        round(new_dimension * voxel_size[1], 3),
                        round(new_dimension * voxel_size[2], 3))

            # Calculate the shifts applied to the chopped map
            if not zero_origin:
                shifts = np.array([0, 0, 0])
            else:
                index_shifts = in_map.index_shifts
                x_shift = (middle_z - radius) * voxel_size[0] + index_shifts[0] * voxel_size[0]
                y_shift = (middle_y - radius) * voxel_size[1] + index_shifts[1] * voxel_size[1]
                z_shift = (middle_x - radius) * voxel_size[2] + index_shifts[2] * voxel_size[2]

                shifts = np.array([x_shift, y_shift, z_shift])

            new_data[np.isnan(new_data)] = 0

            out_map = MapParser('')
            out_map.data = new_data
            out_map.cell = new_cell
            out_map.cellb = in_map.cellb
            out_map.voxel_size = voxel_size
            n_st = in_map.n_start
            if not zero_origin:
                origin = ((middle_z - radius + n_st[0]) * out_map.voxel_size[0] + in_map.origin[0],
                          (middle_y - radius + n_st[1]) * out_map.voxel_size[1] + in_map.origin[1],
                          (middle_x - radius + n_st[2]) * out_map.voxel_size[2] + in_map.origin[2])
                out_map.set_origin(origin)
            chopped_map_obj_list.append(out_map)
            shifts_list.append(shifts)
            # match = self.check_cut_vals(residue, in_map, out_map, shifts)
            # if match:
            # chopped_map_obj_list.append(out_map)
            # shifts_list.append(shifts)
            # else:
            # chopped_map_obj_list.append(None)
            # shifts_list.append(None)
            # print("NO  MATCH")

        if len(shifts_list) > 1 or lst:
            return chopped_map_obj_list, shifts_list
        elif len(shifts_list) == 1:
            return chopped_map_obj_list[0], shifts_list[0]
        else:
            return None, None

    @staticmethod
    def check_cut_vals(model, in_map, out_map, shifts):
        """
        Checks the map values at 5 atoms positions
        :param model:
        :param in_map:
        :param out_map:
        :param shifts:
        :return:
        """
        atoms = [a for a in model.get_atoms()]
        in_coords = []
        out_coords = []

        for i in range(5):
            n = randint(0, len(atoms))
            try:
                in_coords.append(atoms[n].coord)

            except IndexError:
                pass
        for coord in in_coords:
            tmp_coord = list(coord)
            for i in range(3):
                tmp_coord[i] = tmp_coord[i] - shifts[i]
            out_coords.append(tmp_coord)

        in_vals = []
        out_vals = []

        for i in range(len(in_coords)):
            in_indices = in_map.coord_to_index_int(in_coords[i])
            out_indices = out_map.coord_to_index_int(out_coords[i])
            neg_indices = sum(n < 0 for n in in_indices + out_indices)
            if neg_indices > 0:
                in_vals.append(None)
                out_vals.append(None)
            else:
                try:
                    in_vals.append(round(in_map.data[in_indices], 5))
                except IndexError:
                    in_vals.append(None)
                try:
                    out_vals.append(round(out_map.data[out_indices], 5))
                except IndexError:
                    out_vals.append(None)

        if in_vals == out_vals and not all([v is None for v in in_vals]):
            return True
        else:
            print('IN', in_vals)
            print('OU', out_vals)
            return False

    @staticmethod
    def chop_soft_radius(model, in_map, out_map=None, hard_radius=3, soft_radius=2, mask_path=None):
        """
        Chop map using a soft mask with a given radius (hard_radius + soft_radius)
        around the atomic residue. A cosine function is used to create the soft mask.
        :param mask_path: mask output path
        :param model: biopython model object
        :param in_map: path to the input map or strudel map object
        :param out_map: out_map: output map path
        :param hard_radius: hard radius
        :param soft_radius: soft radius, cosine function
        :return strudel map object if out_map not given otherwise None
         """
        # Get atom coordinates
        atoms_coord = []
        for atom in model.get_atoms():
            atoms_coord.append(atom.coord)

        if type(in_map) is str:
            in_map = MapParser(in_map)
        elif type(in_map) is MapParser:
            pass
        else:
            raise TypeError('"in_map" should be a "MapObject" or in_dir to map file')

        voxel_size = in_map.voxel_size
        aver_voxel_size = sum(voxel_size) / 3
        delta1 = hard_radius
        delta2 = hard_radius + soft_radius

        # Create a numpy array for mask
        shape = in_map.data.shape
        mask = np.zeros(shape, dtype='float32')
        for coord in atoms_coord:
            x_index, y_index, z_index = in_map.coord_to_index_int(coord)

            rx = int(round(delta2 / voxel_size[0]))
            ry = int(round(delta2 / voxel_size[1]))
            rz = int(round(delta2 / voxel_size[2]))

            for x in range(x_index - rx, x_index + rx):
                for y in range(y_index - ry, y_index + ry):
                    for z in range(z_index - rz, z_index + rz):
                        # Calculate the distance between the current atom and the current voxel
                        d = aver_voxel_size * math.sqrt((x - x_index) ** 2 + (y - y_index) ** 2 + (z - z_index) ** 2)

                        # Assign mask values based to the distance to the atoms
                        if d <= delta1:
                            try:
                                mask[x, y, z] = 1
                            except IndexError:
                                pass
                        elif delta1 < d < delta2:
                            try:
                                mask[x, y, z] += (math.cos((math.pi / soft_radius) * (d - delta1)) + 1) / 2
                                # if intensity value became > 1 it is set to 1
                                if mask[x, y, z] > 1:
                                    mask[x, y, z] = 1
                            except IndexError:
                                pass

        # Apply the mask to the map data
        final = (mask * in_map.data)

        out_map_obj = MapParser('')
        out_map_obj.copy_header(in_map)
        out_map_obj.data = final

        if mask_path is not None:
            mask_ob = MapParser('')
            mask_ob.copy_header(in_map)
            mask_ob.data = mask
            mask_ob.write_map(mask_path)

        if out_map is not None:
            out_map_obj.write_map(out_map)
        else:
            return out_map_obj

    @staticmethod
    def find_near(atom, model, distance=6):
        close = []
        for atom1 in model.get_atoms():
            if atom1.parent.id != atom.parent.id or atom1.parent.parent.id != atom.parent.parent.id:
                d = atom1 - atom
                if d < distance:
                    close.append(atom1)
        return list(set(close))

    def chop_soft_radius_watershed(self, model, in_map, whole_model, shifts=None, out_map=None,
                                   radius=2, soft_radius=1, mask_path=None):
        """
        Chop map using a soft mask with a given hard_radius around the atomic residue.
        A cosine function is used to create the soft mask. Similar to chop_soft_radius but avoids
        cutting neighboring residues side chains map.
        :param model: biopython atomic residue object
        :param in_map: in_dir to the input map
        :param out_map: out_map: in_dir for the chopped map
        :param radius:
        :param soft_radius:
         """
        # Get atom coordinates
        if shifts is None:
            shifts = np.array([0, 0, 0])

        if type(in_map) is str:
            in_map = MapParser(in_map)
        elif type(in_map) is MapParser:
            pass
        else:
            raise TypeError('"in_map" should be a "MapObject" or in_dir to map file')

        voxel_size = in_map.voxel_size
        aver_voxel_size = sum(voxel_size) / 3

        # Create a numpy array for mask
        shape = in_map.data.shape
        mask = np.zeros(shape, dtype='float32')

        r = int(round((radius + soft_radius) / aver_voxel_size))

        for atom in model.get_atoms():
            xyz = in_map.coord_to_index(atom.coord - shifts)
            xyz_int = in_map.coord_to_index_int(atom.coord - shifts)

            near_atoms = []
            if atom.get_name() not in ['C', 'CA']:
                near_atoms = self.find_near(atom, whole_model, distance=(radius + soft_radius) * 2)

            for x in range(xyz_int[0] - r, xyz_int[0] + r):
                for y in range(xyz_int[1] - r, xyz_int[1] + r):
                    for z in range(xyz_int[2] - r, xyz_int[2] + r):

                        near_ds = [100]
                        for n_atom in near_atoms:
                            n_xyz = in_map.coord_to_index(n_atom.coord - shifts)
                            dn = aver_voxel_size * math.sqrt((x - n_xyz[0]) ** 2 + (y - n_xyz[1]) ** 2
                                                             + (z - n_xyz[2]) ** 2)
                            near_ds.append(dn)
                        dn = min(near_ds)

                        # Calculate the distance between the current atom and the current voxel
                        d = aver_voxel_size * math.sqrt((x - xyz[0]) ** 2 + (y - xyz[1]) ** 2 + (z - xyz[2]) ** 2)
                        if d > dn * 1.3:
                            continue
                        elif dn < radius + soft_radius:
                            delta2 = min((d + dn) * 0.65, radius + soft_radius)
                            delta1 = delta2 - soft_radius
                        else:
                            delta2 = radius + soft_radius
                            delta1 = radius
                        # Assign mask values based to the distance to the atoms
                        if d < delta1:
                            try:
                                mask[x, y, z] = 1
                            except IndexError:
                                pass
                        elif delta1 < d < delta2:
                            try:
                                mask[x, y, z] += (math.cos((math.pi / soft_radius) * (d - delta1)) + 1) / 2
                            except IndexError:
                                pass
        mask[mask > 1] = 1

        final = (mask * in_map.data)

        out_map_obj = MapParser('')
        out_map_obj.copy_header(in_map)
        out_map_obj.data = final

        if mask_path is not None:
            mask_ob = MapParser('')
            mask_ob.copy_header(in_map)
            mask_ob.data = mask
            mask_ob.write_map(mask_path)

        if out_map is not None:
            out_map_obj.write_map(out_map)
        else:
            return out_map_obj

    @staticmethod
    def chop_hard_radius(model, in_map, out_map, radius=3, mask_path=None):
        """
        Chop map using a hard mask with a given hard_radius around the atomic residue.
        :param model: biopython atomic residue object
        :param in_map: in_dir to the input map
        :param out_map: out_map: in_dir for the chopped map
        :param radius:
        """
        # Get atom coordinates
        atoms_coord = []
        for atom in model.get_atoms():
            atoms_coord.append(atom.coord)

        if type(in_map) is str:
            in_map = MapParser(in_map)
        elif type(in_map) is MapParser:
            pass
        else:
            raise TypeError('"in_map" should be a "MapObject" or in_dir to map file')

        voxel_size = in_map.voxel_size
        aver_voxel_size = sum(voxel_size) / 3
        shape = in_map.data.shape
        mask = np.zeros(shape, dtype='float32')

        for coord in atoms_coord:
            x_index, y_index, z_index = in_map.coord_to_index_int(coord)

            rx = int(round(radius / voxel_size[0]))
            ry = int(round(radius / voxel_size[0]))
            rz = int(round(radius / voxel_size[0]))
            for x in range(x_index - rx, x_index + rx):
                for y in range(y_index - ry, y_index + ry):
                    for z in range(z_index - rz, z_index + rz):
                        # Calculate the distance between the current atom and the current voxel
                        d = aver_voxel_size * math.sqrt((x - x_index) ** 2 + (y - y_index) ** 2
                                                        + (z - z_index) ** 2)
                        # Assign mask values based to the distance to the atoms
                        if d < radius:
                            mask[x, y, z] = 1

        # Apply the mask to the map data
        final = (mask * in_map.data)
        # Save the chopped map

        if mask_path is not None:
            mask_ob = MapParser('')
            mask_ob.copy_header(in_map)
            mask_ob.data = mask
            mask_ob.write_map(mask_path)

        out_map_obj = MapParser('')
        out_map_obj.copy_header(in_map)
        out_map_obj.data = final

        if out_map is not None:
            out_map_obj.write_map(out_map)
        else:
            return out_map_obj

    @staticmethod
    def grid_resample_emda(x, num):

        # Forward transform
        X = np.fft.fftn(x)
        X = np.fft.fftshift(X)

        # Placeholder array for output spectrum
        newshape = list(x.shape)
        newshape[0] = num[0]
        newshape[1] = num[1]
        newshape[2] = num[2]
        Y = np.zeros(newshape, X.dtype)

        # upsampling
        if X.shape[0] < newshape[0]:
            dx = abs(newshape[0] - X.shape[0]) // 2
            Y[dx:dx + X.shape[0],
            dx:dx + X.shape[0],
            dx:dx + X.shape[0]] = X
        # downsampling
        if newshape[0] < X.shape[0]:
            dx = abs(newshape[0] - X.shape[0]) // 2
            Y[:, :, :] = X[dx:dx + newshape[0],
                         dx:dx + newshape[0],
                         dx:dx + newshape[0]]

        Y = np.fft.ifftshift(Y)
        return (np.fft.ifftn(Y)).real



"""
mapParser.py

Creates a Map object. This module is built on top of mrcfile library
By default it will transpose the data array so the axis order is x, y, z

Copyright [2013] EMBL - European Bioinformatics Institute
Licensed under the Apache License, Version 2.0 (the
"License"); you may not use this file except in
compliance with the License. You may obtain a copy of
the License at
http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied. See the License for the
specific language governing permissions and limitations
under the License.
"""

__author__ = 'Andrei Istrate'
__email__ = 'andrei@ebi.ac.uk'
__date__ = '2018-05-29'

import os
import mrcfile
import math
import numpy as np


class MapParser:
    def __init__(self, map_path=''):
        self.id = ''
        self.map_path = map_path
        self.data = None
        self.voxel_size = None
        self.n_start = (0, 0, 0)
        self.m = None
        self.cell = None
        self.cellb = (90, 90, 90)
        self.n_xyz = None
        self.origin = (0, 0, 0)
        self.mapc = 1
        self.mapr = 2
        self.maps = 3
        self.info = ''
        self.fraction_matrix = None
        self.index_shifts = (0, 0, 0)
        self.coord_to_index = self._coord_to_index_orthogonal
        self.coord_to_index_int = self._coord_to_index_orthogonal_int
        if os.path.exists(map_path):
            self._load_map()

    def _load_map(self, default_axis=True):
        with mrcfile.open(self.map_path, mode='r+', permissive=True) as mrc:
            self.data = mrc.data
            self.n_start = self._get_n_start(mrc)
            self.m = self._get_m(mrc)
            self.cell = (mrc.header.cella.x, mrc.header.cella.y, mrc.header.cella.z)
            self.cellb = (mrc.header.cellb.alpha, mrc.header.cellb.beta, mrc.header.cellb.gamma)
            self.n_xyz = (mrc.header.nx, mrc.header.ny, mrc.header.nz)
            self.origin = (mrc.header.origin.x, mrc.header.origin.y, mrc.header.origin.z)
            self.id = os.path.basename(self.map_path).split('.')[0]
            self._get_axis_order(mrc)
            if default_axis:
                self._swap_axes()
            self.voxel_size = self._get_voxel_size(mrc)
            self.fraction_matrix = self.calc_fraction_matrix()
            self.index_shifts = self._calc_index_shift()
            if any([z for z in map(lambda k: k != 90, self.cellb)]):
                self.coord_to_index = self._coord_to_index_nonorthogonal
                self.coord_to_index_int = self._coord_to_index_nonorthogonal_int

    def _swap_axes(self):
        crs = (self.mapc, self.mapr, self.maps)
        if crs != (1, 2, 3):
            self.info += 'The input map has non default axis order: '
            order = {1: 'x', 2: 'y', 3: 'z'}
            self.info += 'Changing {} {} {} axis order to x y z'.format(order[crs[0]], order[crs[1]], order[crs[2]])
            # Transpose the data array
            xyz_indices = (crs.index(1), crs.index(2), crs.index(3))
            new_order = [2 - xyz_indices[2 - a] for a in (0, 1, 2)]
            self.data = self.data.transpose(new_order)
            # Rearrange n_start values
            self.n_start = (self.n_start[xyz_indices[0]], self.n_start[xyz_indices[1]], self.n_start[xyz_indices[2]])

            self.mapc = 1
            self.mapr = 2
            self.maps = 3

    @staticmethod
    def _get_voxel_size(mrcfile_object):
        voxel_size = (mrcfile_object.header.cella.x / mrcfile_object.header.mx,
                      mrcfile_object.header.cella.y / mrcfile_object.header.my,
                      mrcfile_object.header.cella.z / mrcfile_object.header.mz)
        return voxel_size

    @staticmethod
    def _get_n_start(mrcfile_object):
        n_start = [int(mrcfile_object.header.nxstart),
                   int(mrcfile_object.header.nystart),
                   int(mrcfile_object.header.nzstart)]
        return n_start

    @staticmethod
    def _get_m(mrcfile_object):
        m = [int(mrcfile_object.header.mx),
             int(mrcfile_object.header.my),
             int(mrcfile_object.header.mz)]
        return m

    def _get_axis_order(self, mrcfile_object):
        self.mapc = int(mrcfile_object.header.mapc)
        self.mapr = int(mrcfile_object.header.mapr)
        self.maps = int(mrcfile_object.header.maps)

    def write_map(self, out_path):
        with mrcfile.new(out_path, overwrite=True) as mr:
            # Assign data and header information
            mr.set_data(self.data)
            mr.update_header_from_data()
            mr.header.cella = self.cell
            mr.header.cellb = self.cellb
            mr.header.nxstart = self.n_start[0]
            mr.header.nystart = self.n_start[1]
            mr.header.nzstart = self.n_start[2]
            try:
                mr.header.mx = self.m[0]
                mr.header.my = self.m[1]
                mr.header.mz = self.m[2]
            except TypeError:
                pass
            mr.header.origin.x = self.origin[0]
            mr.header.origin.y = self.origin[1]
            mr.header.origin.z = self.origin[2]
            mr.header.mapc = self.mapc
            mr.header.mapr = self.mapr
            mr.header.maps = self.maps

    def copy_header(self, map_obj):
        self.voxel_size = map_obj.voxel_size
        self.n_start = map_obj.n_start
        self.cell = map_obj.cell
        self.n_xyz = map_obj.n_xyz
        self.origin = map_obj.origin
        self.cellb = map_obj.cellb

    def _coord_to_index_orthogonal(self, coord):
        """
        Converts atomic residue coordinates into map indices
        :param coord: coordinate in angstrom
        :return: float indices
        """
        x_index = coord[2] / self.voxel_size[2] - self.index_shifts[0]
        y_index = coord[1] / self.voxel_size[1] - self.index_shifts[1]
        z_index = coord[0] / self.voxel_size[0] - self.index_shifts[2]
        return x_index, y_index, z_index

    def _coord_to_index_orthogonal_int(self, coord):
        """
        Converts atomic residue coordinates into map indices
        :param coord: coordinate in angstrom
        :return: int indices
        """
        x, y, z = self._coord_to_index_orthogonal(coord)
        return int(round(x)), int(round(y)), int(round(z))

    def _coord_to_index_nonorthogonal(self, coord):
        coord = np.flip(coord)
        return self.fraction_matrix.dot(coord)*self.n_xyz - self.index_shifts

    def _coord_to_index_nonorthogonal_int(self, coord):
        x, y, z = self._coord_to_index_nonorthogonal(coord)
        return int(round(x)), int(round(y)), int(round(z))

    def calc_fraction_matrix(self):
        """
        Calculates matrix used to transform Cartesian
        coordinates in the ATOM_SITE category to fractional coordinates
        in the same category
        :return: fractional matrix
        """
        cell = self.cell
        ang = self.cellb
        ang = (ang[0] * math.pi / 180, ang[1] * math.pi / 180, ang[2] * math.pi / 180)

        omega = cell[0] * cell[1] * cell[2] * math.sqrt(
            1 - math.cos(ang[0]) ** 2 - math.cos(ang[1]) ** 2 - math.cos(ang[2]) ** 2 +
            2 * math.cos(ang[0]) * math.cos(ang[1]) * math.cos(ang[0]))
        m11 = 1 / cell[0]
        m12 = - math.cos(ang[2]) / (cell[0] * math.sin(ang[2]))
        m13 = cell[1] * cell[2] * (math.cos(ang[0]) * math.cos(ang[2]) - math.cos(ang[1])) / (omega * math.sin(ang[2]))

        m21 = 0
        m22 = 1 / (cell[1] * math.sin(ang[2]))
        m23 = cell[0] * cell[2] * (math.cos(ang[1]) * math.cos(ang[2]) - math.cos(ang[0])) / (omega * math.sin(ang[2]))

        m31 = 0
        m32 = 0
        m33 = cell[0] * cell[1] * math.sin(ang[2]) / omega
        matrix = [m11, m12, m13, m21, m22, m23, m31, m32, m33]

        return np.array(matrix).reshape(3, 3)

    def _calc_index_shift_(self):
        if self.origin == (0, 0, 0):
            x = self.n_start[0]
            y = self.n_start[1]
            z = self.n_start[2]
        elif self.n_start != (0, 0, 0):
            x = self.origin[2] / self.voxel_size[2]
            y = self.origin[1] / self.voxel_size[1]
            z = self.origin[0] / self.voxel_size[0]
        else:
            x, y, z, = 0, 0, 0
        return x, y, z

    def _calc_index_shift(self):
        if self.origin == (0, 0, 0):
            x = self.n_start[0]
            y = self.n_start[1]
            z = self.n_start[2]
        else: #self.n_start != (0, 0, 0):
            x = self.origin[2] / self.voxel_size[2]
            y = self.origin[1] / self.voxel_size[1]
            z = self.origin[0] / self.voxel_size[0]
        # else:
        #     x, y, z, = 0, 0, 0
        return x, y, z

    def set_origin(self, origin):
        self.origin = origin
        self.index_shifts = self._calc_index_shift()

    def grid_resample_emda(self, target_voxel):
        x = self.data

        if x.shape[0] % 2 != 0:
            xshape = list(x.shape)
            xshape[0] = xshape[0] + 1
            xshape[1] = xshape[1] + 1
            xshape[2] = xshape[2] + 1
            temp = np.zeros(xshape, x.dtype)
            temp[:-1, :-1, :-1] = x
            x = temp
            self.cell = (self.cell[0] + self.voxel_size[0], self.cell[1] + self.voxel_size[1], self.cell[2] + self.voxel_size[2])

        scale = self.voxel_size[0] / target_voxel
        old_grid = x.data.shape
        new_grid = [int(round(i * scale)) for i in old_grid]

        # Forward transform
        X = np.fft.fftn(x)
        X = np.fft.fftshift(X)
        # Placeholder array for output spectrum
        newshape = list(x.shape)

        if new_grid[0] % 2 != 0:
            newshape[0] = new_grid[0] + 1
            newshape[1] = new_grid[1] + 1
            newshape[2] = new_grid[2] + 1
        else:
            newshape[0] = new_grid[0]
            newshape[1] = new_grid[1]
            newshape[2] = new_grid[2]

        if x.shape[0] == newshape[0]:
            return x
        Y = np.zeros(newshape, X.dtype)
        # upsampling
        if X.shape[0] < newshape[0]:
            dx = abs(newshape[0] - X.shape[0]) // 2
            Y[dx:dx + X.shape[0],
            dx:dx + X.shape[0],
            dx:dx + X.shape[0]] = X
        # downsampling
        if newshape[0] < X.shape[0]:
            dx = abs(newshape[0] - X.shape[0]) // 2
            Y[:, :, :] = X[dx:dx + newshape[0],
                         dx:dx + newshape[0],
                         dx:dx + newshape[0]]

        Y = np.fft.ifftshift(Y)

        self.data = np.float32((np.fft.ifftn(Y)).real)
        self.m = self.data.shape
        self.voxel_size = (self.cell[0]/self.m[0], self.cell[1]/self.m[1], self.cell[2]/self.m[2])
        self.index_shifts = self._calc_index_shift()




"""
functions.py

General purpose functions

Copyright [2013] EMBL - European Bioinformatics Institute
Licensed under the Apache License, Version 2.0 (the
"License"); you may not use this file except in
compliance with the License. You may obtain a copy of
the License at
http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied. See the License for the
specific language governing permissions and limitations
under the License.
"""

__author__ = 'Andrei Istrate'
__email__ = 'andrei@ebi.ac.uk'
__date__ = '2019-01-29'

import math
import logging
from datetime import datetime, timedelta
import time
import json
import mrcfile


def global_except_hook(exctype, value, traceback):
    import sys
    try:
        import mpi4py.MPI
        sys.stderr.write("\n*****************************************************\n")
        sys.stderr.write("Uncaught exception was detected on rank {}. \n".format(
            mpi4py.MPI.COMM_WORLD.Get_rank()))
        from traceback import print_exception
        print_exception(exctype, value, traceback)
        sys.stderr.write("*****************************************************\n\n")
        sys.stderr.write("\n")
        sys.stderr.write("Calling MPI_Abort() to shut down MPI processes...\n")
        sys.stderr.flush()
    finally:
        try:
            import mpi4py.MPI
            mpi4py.MPI.COMM_WORLD.Abort(1)
        except Exception as e:
            sys.stderr.write("*****************************************************\n")
            sys.stderr.write("Sorry, we failed to stop MPI, this process will hang.\n")
            sys.stderr.write("*****************************************************\n")
            sys.stderr.flush()
            raise e


def read_json(json_path):
    with open(json_path) as j:
        data = json.load(j)
    return data


def save_json(file_path, data):
    with open(file_path, 'w') as j:
        json.dump(data, j, indent=4)


def split_list(inp_list, nr):
    """
    Splits evenly a list
    :param inp_list: list
    :param nr: number of parts
    :return: list of "nr" lists
    """
    new_list = []
    nr_el = 1.0 / nr * len(inp_list)
    for i in range(nr):
        start = int(round(i * nr_el))
        end = int(round((i + 1) * nr_el))
        new_list.append(inp_list[start:end])
    return new_list


def split_sorted_list(inp_list, nr):
    new_list = []

    for i in range(nr):
        tmp = []
        for k in range(i, len(inp_list), nr):
            try:
                tmp.append(inp_list[k])
            except IndexError:
                pass
        new_list.append(tmp)
    return new_list


def split_weighted_dict(dictionary, nr):
    """
    From a dictionary ({'ASP': 20, 'GLU': 30, 'PHE': 10, 'ARG': 50, ..)
    Creates a list of lists of keys of the input dictionary such as in each
    group the sum of the corresponding values is minimal.
    :param dictionary:
    :param nr:
    :return:
    """
    out_list = []
    weights = []
    sorted_dict = sorted(dictionary.items(), key=lambda kv: kv[1], reverse=True)

    for key, value in sorted_dict[0:nr]:
        out_list.append([key])
        weights.append(value)

    for key, value in sorted_dict[nr:]:
        min_index = weights.index(min(weights))
        out_list[min_index].append(key)
        weights[min_index] = weights[min_index] + value

    return out_list


def setup_logger(name, log_file=None, warning_level="info"):
    """
    Function setup as many loggers as needed
    :param name: logger name
    :param log_file: log file in_dir
    :param warning_level: warning level
    :return: logger
    """
    if warning_level.lower() == 'debug':
        warning_level = logging.DEBUG
    else:
        warning_level = logging.INFO
    formatter = logging.Formatter('%(message)s')
    if log_file:
        handler = logging.FileHandler(log_file)
    else:
        handler = logging.StreamHandler()
    handler.setFormatter(formatter)

    logger = logging.getLogger(name)
    logger.setLevel(warning_level)
    # Remove old handlers
    for h in logger.handlers:
        logger.removeHandler(h)
    logger.addHandler(handler)

    return logger


def report_elapsed(start):
    """
    Reports elapsed time since start in human readable formatted
    :param start: script starting time
    :return: report
    """
    end = time.time()
    elapsed = round(end - start, 2)
    sec = timedelta(seconds=int(elapsed))
    d = datetime(1, 1, 1) + sec
    formatted = "d:h:m:s {}:{}:{}:{}".format(d.day - 1, d.hour, d.minute, d.second)
    text = '{} seconds   {}'.format(elapsed, formatted)
    return text


def print_time():
    """
    Returns human readable date and time
    :return:
    """
    date_time = datetime.now().strftime("%Y-%m-%d  %H:%M:%S")
    return date_time


def find_good_grid(in_grid):
    def largest_prime_factor(n):
        i = 2
        while i * i <= n:
            if n % i:
                i += 1
            else:
                n //= i
        return n

    out_grid = []
    for dimension in in_grid:
        new_dimension = dimension
        if new_dimension % 2:
            new_dimension += 1
        largest_prime = largest_prime_factor(new_dimension)
        while largest_prime > 19:
            new_dimension += 2
            largest_prime = largest_prime_factor(new_dimension)
        out_grid.append(new_dimension)
    return tuple(out_grid)


def get_map_grid_size(map_path):
    with mrcfile.mmap(map_path, mode='r+', permissive=True) as mrc:
        grid = (mrc.header.mx, mrc.header.my, mrc.header.mz)
    return grid


def calc_fraction_matrix(cell, ang):
    """
    Calculates matrix used to transform Cartesian
    coordinates in the ATOM_SITE category to fractional coordinates
    in the same category
    :param cell: unit cell
    :param ang: unit cell angles in degrees
    :return:
    """
    ang = (ang[0]*math.pi/180, ang[1]*math.pi/180, ang[2]*math.pi/180)

    omega = cell[0]*cell[1]*cell[2] * math.sqrt(1 - math.cos(ang[0])**2 - math.cos(ang[1])**2 - math.cos(ang[2])**2 +
                                                2*math.cos(ang[0])*math.cos(ang[1])*math.cos(ang[0]))
    m11 = 1 / cell[0]
    m12 = - math.cos(ang[2]) / (cell[0]*math.sin(ang[2]))
    m13 = cell[1]*cell[2] * (math.cos(ang[0])*math.cos(ang[2]) - math.cos(ang[1])) / (omega * math.sin(ang[2]))

    m21 = 0
    m22 = 1 / (cell[1]*math.sin(ang[2]))
    m23 = cell[0]*cell[2] * (math.cos(ang[1])*math.cos(ang[2]) - math.cos(ang[0])) / (omega * math.sin(ang[2]))

    m31 = 0
    m32 = 0
    m33 = cell[0]*cell[1]*math.sin(ang[2]) / omega
    matrix = [m11, m12, m13, m21, m22, m23, m31, m32, m33]
    # for i, el in enumerate(matrix):
    #     matrix[i] = round(el, 6)

    return matrix


